# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage defines region shapes.
"""
from .annulus import *  # noqa
from .circle import *  # noqa
from .ellipse import *  # noqa
from .line import *  # noqa
from .point import *  # noqa
from .polygon import *  # noqa
from .rectangle import *  # noqa
from .text import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage provides tools for reading and writing region files.
"""
from .core import *  # noqa

from .ds9.connect import *  # noqa
from .ds9.core import *  # noqa
from .ds9.read import *  # noqa
from .ds9.write import *  # noqa

from .crtf.connect import *  # noqa
from .crtf.core import *  # noqa
from .crtf.read import *  # noqa
from .crtf.write import *  # noqa

from .fits.connect import *  # noqa
from .fits.core import *  # noqa
from .fits.read import *  # noqa
from .fits.write import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Regions is Astropy coordinated package to provide tools for region
handling.
"""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *  # noqa
# ----------------------------------------------------------------------------

from .core import *  # noqa
from .io import *  # noqa
from .shapes import *  # noqa
from ._utils.examples import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pickle
import tempfile

from astropy.coordinates import SkyCoord
import astropy.units as u
import pytest

from ....shapes.ellipse import EllipseSkyRegion
from ..write import _write_crtf

try:
    from casatools import image, simulator
    from casatasks import tclean
    HAS_CASATOOLS = True
except ImportError:
    HAS_CASATOOLS = False


@pytest.mark.skipif('not HAS_CASATOOLS')
def test_casa_masking():
    with tempfile.TemporaryDirectory() as tmpdir:
        # SIMULATE SOME DATA SET

        # Define antennas
        diam = [25, 25, 25, 25, 25]
        xx = [50, 100, 150, 200, 250]
        yy = [2, -5, -20, -50, -100]
        zz = [-0.5, -1.0, -1.5, -2.0, -2.5]

        sm = simulator()
        sm.open(tmpdir + '/SIM.ms')
        # do configuration
        posvla = me.observatory('VLA')
        sm.setconfig(telescopename='VLA', x=xx, y=yy, z=zz, dishdiameter=diam,
                     mount='alt-az', antname='VLA',
                     coordsystem='local', referencelocation=posvla)

        # Initialize the spectral windows
        sm.setspwindow(spwname='CBand', freq='5GHz', deltafreq='50MHz',
                       freqresolution='50MHz', nchannels=1,
                       stokes='RR RL LR LL')

        # Initialize the source and calibrater
        sm.setfield(sourcename='My cal',
                    sourcedirection=['J2000', '00h0m0.0', '+45.0.0.000'],
                    calcode='A')
        sm.setfield(sourcename='My source',
                    sourcedirection=['J2000', '01h0m0.0', '+47.0.0.000'])

        sm.setlimits(shadowlimit=0.001, elevationlimit='8.0deg')
        sm.setauto(autocorrwt=0.0)

        sm.settimes(integrationtime='10s', usehourangle=False,
                    referencetime=me.epoch('utc', 'today'))

        sm.observe('My cal', 'CBand', starttime='720s', stoptime='1020s')
        sm.observe('My source', 'CBand', starttime='1030s', stoptime='1500s')

        sm.close()

        # Create mask to use during clean
        reg = EllipseSkyRegion(SkyCoord(0.0 * u.deg, 45.0 * u.deg,
                                        frame='fk5'),
                               width=1.0 * u.arcmin, height=2.0 * u.arcmin,
                               angle=45 * u.deg)
        _write_crtf([reg], tmpdir + '/SIM.crtf', 'fk5', '.6f', 'deg')

        # Image the dataset
        tclean(vis=tmpdir + '/SIM.ms', imagename=tmpdir + '/SIM', imsize=100,
               cell='5arcsec', niter=1, mask=tmpdir + '/SIM.crtf',
               interactive=False)

        ia = image()
        ia.open(tmpdir + '/SIM.mask')
        mask_array = ia.getregion()
        ia.close()

        with open('data/binary_mask.pkl', 'rb') as f:
            ref_mask = pickle.load(f)

        assert all(mask_array == ref_mask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides a Regions class.
"""

from .core import Region
from .registry import RegionsRegistry

__all__ = ['Regions']
__doctest_skip__ = ['Regions.read', 'Regions.write', 'Regions.parse',
                    'Regions.serialize']


class Regions:
    """
    Class to hold a list of `~regions.Region` objects.

    This class provides a unified I/O interface that supports reading,
    writing, parsing, and serializing many region data formats.

    Parameters
    ----------
    regions : list of `~regions.Region`
        The list of region objects.
    """

    def __init__(self, regions):
        self.regions = regions

    def __getitem__(self, index):
        newregions = self.regions[index]
        if isinstance(newregions, Region):  # one item
            return newregions
        else:
            newcls = object.__new__(self.__class__)
            newcls.regions = newregions
            return newcls

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f'<{cls_name}({repr(self.regions)})>'

    def __str__(self):
        return str(self.regions)

    def __len__(self):
        return len(self.regions)

    def append(self, region):
        """
        Append the region to the end of the list of regions.

        Parameters
        ----------
        region : `~regions.Region`
            The region to append.
        """
        self.regions.append(region)

    def extend(self, regions):
        """
        Extend the list of regions by appending elements from the
        input regions.

        Parameters
        ----------
        regions : list of `~regions.Region`
            A list of regions to include.
        """
        self.regions.extend(regions)

    def insert(self, index, region):
        """
        Insert the region before index.

        Parameters
        ----------
        index : int
            The list index.
        region : `~regions.Region`
            The region to insert.
        """
        self.regions.insert(index, region)

    def reverse(self):
        """
        Reverse the list of regions in place.
        """
        self.regions.reverse()

    def pop(self, index=-1):
        """
        Remove and return the region at index.

        Parameters
        ----------
        index : int, optional
            The index of the region to remove.

        Returns
        -------
        result : `~regions.Region`
        """
        return self.regions.pop(index)

    def copy(self):
        """
        Return a shallow copy of this object.
        """
        newcls = object.__new__(self.__class__)
        newcls.regions = self.regions.copy()
        return newcls

    @classmethod
    def get_formats(cls):
        """
        Get the registered I/O formats as a Table.
        """
        return RegionsRegistry.get_formats(cls)

    @classmethod
    def read(cls, filename, format=None, cache=False, **kwargs):
        """
        Read and parse a region file and return as a Regions object.

        This method allows reading a file in many supported data
        formats, e.g.,::

            >>> from regions import Regions
            >>> reg1 = Regions.read('regions.reg', format='ds9')
            >>> reg2 = Regions.read('regions.crtf', format='crtf')
            >>> reg3 = Regions.read('regions.fits', format='fits')

        A list of the available formats for `~regions.Regions` is
        available using::

            >>> Regions.get_formats()

        Parameters
        ----------
        filename : str
            The filename or URL of the file to read.

        format : str, optional
            The file format specifier.

        cache : bool or 'update', optional
            Whether to cache the contents of remote URLs. If 'update',
            check the remote URL for a new version but store the result
            in the cache.

        **kwargs : dict, optional
            Keyword arguments passed to the data reader.

        Returns
        -------
        result : `~regions.Regions`
            A `~regions.Regions` object containing the file contents.
        """
        return RegionsRegistry.read(filename, cls, format=format,
                                    cache=cache, **kwargs)

    @classmethod
    def parse(cls, data, format=None, **kwargs):
        """
        Parse a region string or table and return as a Regions object.

        This method allows parsing region data in many supported data
        formats, e.g.,::

            >>> from regions import Regions
            >>> reg1 = Regions.parse(regions_str, format='ds9')
            >>> reg2 = Regions.parse(regions_str, format='crtf')
            >>> reg3 = Regions.parse(regions_tbl, format='fits')

        A list of the available formats for `~regions.Regions` is
        available using::

            >>> Regions.get_formats()

        Parameters
        ----------
        data : str or `~astropy.table.Table`
            The region data to parse.

        format : str, optional
            The file format specifier.

        **kwargs : dict, optional
            Keyword arguments passed to the data parser.

        Returns
        -------
        result : `~regions.Regions`
            A `~regions.Regions` object containing the data contents.
        """
        return RegionsRegistry.parse(data, cls, format=format,
                                     **kwargs)

    def write(self, filename, format=None, overwrite=False, **kwargs):
        """
        Write the regions to a region file in the specified format.

        This method allows writing a file in many supported data
        formats, e.g.,::

            >>> from regions import Regions
            >>> reg = Regions.read('regions.reg', format='ds9')
            >>> reg.write('new_regions.reg', format='ds9')
            >>> reg.write('new_regions.crtf', format='crtf')
            >>> reg.write('new_regions.fits', format='fits')

        A list of the available formats for `~regions.Regions` is
        available using::

            >>> Regions.get_formats()

        Parameters
        ----------
        filename : str
            The filename or URL of the file to write.

        format : str, optional
            The file format specifier.

        overwrite : bool, optional
            If True, overwrite the output file if it exists. Raises an
            `OSError` if False and the output file exists. Default is
            False.

        **kwargs : dict, optional
            Keyword arguments passed to the data writer.
        """
        return RegionsRegistry.write(self.regions, filename,
                                     self.__class__, format=format,
                                     overwrite=overwrite, **kwargs)

    def serialize(self, format=None, **kwargs):
        """
        Serialize the regions to a region string or table.

        This method allows serializing regions in many supported data
        formats, e.g.,::

            >>> from regions import Regions
            >>> reg = Regions.read('regions.reg', format='ds9')
            >>> reg1_str = reg.serialize(format='ds9')
            >>> reg2_str = reg.serialize(format='crtf')
            >>> reg3_tbl = reg.serialize(format='fits')

        A list of the available formats for `~regions.Regions` is
        available using::

            >>> Regions.get_formats()

        Parameters
        ----------
        format : str, optional
            The file format specifier.

        **kwargs : dict, optional
            Keyword arguments passed to the data serializer.
        """
        return RegionsRegistry.serialize(self.regions, self.__class__,
                                         format=format, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import copy

from astropy.coordinates import SkyCoord
import numpy as np

__all__ = ['PixCoord']


# Define global variables for the default 'origin' and 'mode' used for
# WCS transformations.
_DEFAULT_WCS_ORIGIN = 0
_DEFAULT_WCS_MODE = 'all'


class PixCoord:
    """
    A class for pixel coordinates.

    This class can represent a scalar or an array of pixel coordinates.

    The data members are either numbers or `~numpy.ndarray`
    (not `~astropy.units.Quantity` objects with unit "pixel").

    Given a `astropy.wcs.WCS` object, it can be transformed to and from
    a `~astropy.coordinates.SkyCoord` object.

    Parameters
    ----------
    x : float or array-like
        Pixel coordinate x value.
    y : float or array-like
        Pixel coordinate y value.

    Examples
    --------
    Usage examples are provided in the :ref:`getting_started-coord`
    section of the documentation.
    """

    def __init__(self, x, y):
        x, y = np.broadcast_arrays(x, y)

        if x.shape == ():
            self.x, self.y = x.item(), y.item()
        else:
            self.x, self.y = x, y

    def copy(self):
        return self.__class__(copy.deepcopy(self.x), copy.deepcopy(self.y))

    @staticmethod
    def _validate(val, name, expected='any'):
        """
        Validate that a given object is an appropriate `PixCoord`.

        This is used for input validation throughout the regions
        package, especially in the ``__init__`` method of pixel region
        classes.

        Parameters
        ----------
        val : `PixCoord`
            The object to check.
        name : str
            Parameter name (used for error messages).

        expected : {'any', 'scalar', 'not scalar'}
            What kind of PixCoord to check for.

        Returns
        -------
        val : `PixCoord`
            The input object (at the moment unmodified, might do fix-ups
            here later).
        """
        if not isinstance(val, PixCoord):
            raise TypeError(f'{name} must be a PixCoord')

        if expected == 'any':
            pass
        elif expected == 'scalar':
            if not val.isscalar:
                raise ValueError(f'{name} must be a scalar PixCoord')
        elif expected == 'not scalar':
            if val.isscalar:
                raise ValueError(f'{name} must be a non-scalar PixCoord')
        else:
            raise ValueError(f'Invalid argument for `expected`: {expected}')

        return val

    @property
    def isscalar(self):
        """
        Whether the instance is scalar (e.g., a single (x, y)
        coordinate).
        """
        return np.isscalar(self.x)

    def __repr__(self):
        return f'{self.__class__.__name__}(x={self.x}, y={self.y})'

    def __len__(self):
        if self.isscalar:
            raise TypeError(f'Scalar {self.__class__.__name__!r} object has '
                            'no len()')
        return len(self.x)

    def __iter__(self):
        for (x, y) in zip(self.x, self.y):
            yield PixCoord(x=x, y=y)

    def __getitem__(self, key):
        if self.isscalar:
            raise IndexError(f'Scalar {self.__class__.__name__!r} cannot be '
                             'indexed or sliced.')

        # Let Numpy do the slicing
        x = self.x[key]
        y = self.y[key]
        return PixCoord(x=x, y=y)

    def __eq__(self, other):
        """
        Checks whether ``other`` is `PixCoord` object and whether
        their abscissa and ordinate values are equal using
        `np.testing.assert_allclose` with its default tolerance values.
        """
        if isinstance(other, PixCoord):
            return np.allclose([self.x, self.y], [other.x, other.y])
        else:
            return False

    def to_sky(self, wcs, origin=_DEFAULT_WCS_ORIGIN, mode=_DEFAULT_WCS_MODE):
        """
        Convert to a `~astropy.coordinates.SkyCoord`.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            The WCS to use to convert pixels to world coordinates.
        origin : int, optional
            Whether to return 0 or 1-based pixel coordinates.
        mode : {'all', 'wcs'}, optional
            Whether to do the transformation including distortions
            (``'all'``) or only including only the core WCS
            transformation (``'wcs'``).

        Returns
        -------
        coord : `~astropy.coordinates.SkyCoord`
            A new object with sky coordinates corresponding to the pixel
            coordinates.
        """
        return SkyCoord.from_pixel(xp=self.x, yp=self.y, wcs=wcs,
                                   origin=origin, mode=mode)

    @classmethod
    def from_sky(cls, skycoord, wcs, origin=_DEFAULT_WCS_ORIGIN,
                 mode=_DEFAULT_WCS_MODE):
        """
        Create `PixCoord` from a `~astropy.coordinates.SkyCoord`.

        Parameters
        ----------
        skycoord : `~astropy.coordinates.SkyCoord`
            The sky coordinate.
        wcs : `~astropy.wcs.WCS`
            The WCS to use to convert pixels to world coordinates.
        origin : int, optional
            Whether to return 0 or 1-based pixel coordinates.
        mode : {'all', 'wcs'}, optional
            Whether to do the transformation including distortions
            (``'all'``) or only including only the core WCS
            transformation (``'wcs'``).

        Returns
        -------
        coord : `PixCoord`
            A new `PixCoord` object at the position of the input sky
            coordinates.
        """
        x, y = skycoord.to_pixel(wcs=wcs, origin=origin, mode=mode)
        return cls(x=x, y=y)

    def separation(self, other):
        r"""
        Calculate the separation to another pixel coordinate.

        This is the two-dimensional Cartesian separation :math:`d` where

        .. math::
            d = \sqrt{(x_1 - x_2) ^ 2 + (y_1 - y_2) ^ 2}

        Parameters
        ----------
        other : `PixCoord`
            The other pixel coordinate.

        Returns
        -------
        separation : `numpy.array`
            The separation in pixels.
        """
        dx = other.x - self.x
        dy = other.y - self.y
        return np.hypot(dx, dy)

    @property
    def xy(self):
        """
        A 2-tuple ``(x, y)`` for this coordinate.
        """
        return self.x, self.y

    def rotate(self, center, angle):
        """
        Rotate the pixel coordinate.

        Postive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        coord : `PixCoord`
            The rotated coordinates (which is an independent copy).
        """
        dx = self.x - center.x
        dy = self.y - center.y
        vec = np.array([dx, dy])

        cosa, sina = np.cos(angle), np.sin(angle)
        rotation_matrix = np.array([[cosa, -sina], [sina, cosa]])

        vec = np.matmul(rotation_matrix, vec)

        return self.__class__(center.x + vec[0], center.y + vec[1])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from warnings import warn

from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.utils.decorators import deprecated
from astropy.wcs import WCS
import numpy as np

from ...core import Regions
from ...core.registry import RegionsRegistry
from ..core import _Shape, _ShapeList, reg_mapping
from .core import (FITSRegionParserError, FITSRegionParserWarning,
                   language_spec)

__all__ = ['FITSRegionParser', 'read_fits']


class _FITSRegionParser:
    """
    Parses a FITS Region table.

    Parameters
    ----------
    table : `~astropy.table.Table`
        A FITS region table.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """

    valid_columns = ['X', 'Y', 'SHAPE', 'COMPONENT', 'R', 'ROTANG']

    def __init__(self, table, errors='strict'):
        if errors not in ('strict', 'ignore', 'warn'):
            raise ValueError('errors must be one of "strict", "ignore", or '
                             '"warn"')
        if not isinstance(table, Table):
            raise TypeError('The table should be an astropy table object')
        self.table = table
        self.errors = errors
        self.unit = {}
        self._shapes = {}

        self.parse_table()

    @property
    def shapes(self):
        shape_list = _ShapeList()
        components = list(self._shapes.keys())
        components.sort()
        for component in components:
            shape_list += _ShapeList(self._shapes[component])
        return shape_list

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, FITSRegionParserWarning)
        elif self.errors == 'strict':
            raise FITSRegionParserError(msg)

    def parse_table(self):
        for col in self.table.colnames:
            if col not in self.valid_columns:
                self._raise_error('This table has an invalid column name: '
                                  f'"{col}"')
            else:
                self.unit[col] = self.table[col].unit

        if 'COMPONENT' not in self.table.colnames:
            self.table['COMPONENT'] = np.ones([1, len(self.table)])

        for row in self.table:
            reg = _FITSRegionRowParser(row, self.unit, self.errors)
            component, shape = reg.parse()
            if component in self.shapes:
                self._shapes[component].append(shape)
            else:
                self._shapes[component] = [shape]


class _FITSRegionRowParser():
    """
    Parse a single row of a FITS region table.

    Parameters
    ----------
    row : `~astropy.table.Row`
        A row of the region table that is to be parsed.

    unit : dict
        The units of each column in the row.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do
        nothing (i.e., be silent).
    """

    def __init__(self, row, unit, errors='strict'):
        self.errors = errors
        self.unit = unit
        self.row = row

        region_type = self._get_col_value('SHAPE0', 'POINT')[0]

        # Default region is POINT
        if region_type == '':
            region_type = 'POINT'

        if region_type[0] == '!':
            self.include = False
            region_type = region_type[1:]
        else:
            self.include = True

        region_type = region_type.strip().upper()

        if region_type in language_spec:
            self.region_type = region_type
        else:
            self._raise_error(f'"{region_type}" is not a valid FITS Region '
                              'type')

        self.component = str(self.row['COMPONENT'])

    def _get_col_value(self, colname, default=None):
        index = None

        if colname[-1].isdigit():
            index = int(colname[-1])
            colname = colname[: -1]

        try:
            val, unit = self.row[colname], self.unit[colname]

            if np.isscalar(val):
                val = np.array(val).reshape(1, )
            if index is not None:
                if index < len(val) and val[index] is not None:
                    return val[index], unit
                else:
                    raise ValueError(f'The column "{colname}" must have more '
                                     f'than {index} value for the region '
                                     f'{self.region_type}')
            else:
                return val, unit
        except KeyError:
            if default is None:
                self._raise_error(f'The column "{colname}" is missing in '
                                  'the table')
            else:
                return default

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, FITSRegionParserWarning)
        elif self.errors == 'strict':
            raise FITSRegionParserError(msg)

    def parse(self):
        coords = []

        for x in language_spec[self.region_type]:
            y, unit = self._get_col_value(x)
            coords.append(self._parse_value(y, unit))

        meta = {'tag': self.component}

        if self.region_type == 'POLYGON':
            coords_new = []
            for x, y in zip(coords[0], coords[1]):
                coords_new += [x, y]
            coords = coords_new
        elif self.region_type == 'BOX':
            # Add a 0-degree rotation to turn it into ROTBOX
            coords.append(0.0 * u.deg)

        region_type = self.region_type.lower()
        if region_type in reg_mapping['FITS_REGION']:
            region_type = reg_mapping['FITS_REGION'][region_type]
        else:
            self._raise_error(f'"{self.region_type}" is currently not '
                              'supported')

        return self.component, _Shape('physical', region_type,
                                      coords, meta, False, False)

    def _parse_value(self, val, unit):
        units = dict(pix=u.dimensionless_unscaled,
                     deg=u.deg, rad=u.rad)
        if unit is None:
            self._raise_error(f'The unit: {unit} is invalid')

        return val * units.get(str(unit), unit)


@RegionsRegistry.register(Regions, 'read', 'fits')
def _read_fits(filename, errors='strict', cache=False):
    """
    Read a FITS region file, converting a FITS regions table to a list
    of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The file path.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do
        nothing (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    with fits.open(filename, cache=cache) as hdul:
        sky_regions = []
        for hdu in hdul:
            if hdu.name == 'REGION':
                region_table = Table.read(hdu)
                regions = _parse_fits(region_table, errors=errors)

                wcs = WCS(hdu.header, keysel=['image', 'binary', 'pixel'])
                for reg in regions:
                    sky_regions.append(reg.to_sky(wcs))
        return Regions(sky_regions)


@RegionsRegistry.register(Regions, 'parse', 'fits')
def _parse_fits(region_table, errors='strict'):
    parser = _FITSRegionParser(region_table, errors=errors)
    return Regions(parser.shapes.to_regions())


@deprecated('0.5', alternative='`regions.Regions.read`')
def read_fits_region(filename, errors='strict', cache=False):
    """
    Read a FITS region file, converting a FITS regions table to a list
    of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The file path.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do
        nothing (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    return _read_fits(filename, errors=errors, cache=cache)


@deprecated('0.5', alternative='`regions.Regions.read`')
def read_fits(filename, errors='strict', cache=False):
    """
    Read a FITS region file, converting a FITS regions table to a list
    of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The file path.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do
        nothing (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    return _read_fits(filename, errors=errors, cache=cache)


@deprecated('0.5', alternative='`regions.Regions.parse`')
class FITSRegionParser(_FITSRegionParser):
    """
    Parses a FITS Region table.

    Parameters
    ----------
    table : `~astropy.table.Table`
        A FITS region table.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.FITSRegionParserError`. 'warn' will raise a
        `~regions.FITSRegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This file sets up detailed tests for computing masks with reference images.
"""
import itertools

import astropy.units as u
import pytest

from ...core import PixCoord
from ...shapes.circle import CirclePixelRegion
from ...shapes.ellipse import EllipsePixelRegion
from ...shapes.rectangle import RectanglePixelRegion
from ...shapes.polygon import PolygonPixelRegion

REGIONS = [CirclePixelRegion(PixCoord(3.981987, 4.131378), radius=3.3411),
           EllipsePixelRegion(PixCoord(5.981987, 4.131378), width=10.4466,
                              height=6.6822, angle=12 * u.deg),
           RectanglePixelRegion(PixCoord(5.981987, 4.131378), width=7.2233,
                                height=4.3411, angle=12 * u.deg),
           PolygonPixelRegion(PixCoord([-2.334, 3.631, 1.122, -1.341],
                                       [-3.121, -2.118, 2.987, 1.759]))]

MODES = [{'mode': 'center'},
         {'mode': 'exact'},
         {'mode': 'subpixels', 'subpixels': 1},
         {'mode': 'subpixels', 'subpixels': 5},
         {'mode': 'subpixels', 'subpixels': 18}]


def label(value):
    if isinstance(value, CirclePixelRegion):
        return 'circ'
    elif isinstance(value, EllipsePixelRegion):
        return 'elli'
    elif isinstance(value, RectanglePixelRegion):
        return 'rect'
    elif isinstance(value, PolygonPixelRegion):
        return 'poly'
    else:
        return '-'.join(f'{key}_{value}'
                        for key, value in sorted(value.items()))


@pytest.mark.array_compare(fmt='text', write_kwargs={'fmt': '%12.8e'})
@pytest.mark.parametrize(('region', 'mode'),
                         itertools.product(REGIONS, MODES), ids=label)
def test_to_mask(region, mode):
    try:
        mask = region.to_mask(**mode)
    except NotImplementedError:
        pytest.xfail()
    return mask.data.astype(float)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module provides a RegionsRegistry class.
"""

from astropy.table import Table

__all__ = []


class IORegistryError(Exception):
    """Exception class for various registry errors."""


class RegionsRegistry:
    """
    Class to hold a registry to read, write, parse, and serialize regions
    in various formats.
    """

    registry = {}

    @classmethod
    def register(cls, classobj, methodname, filetype):
        def inner_wrapper(wrapped_func):
            key = (classobj, methodname, filetype)
            if key in cls.registry:
                raise ValueError(f'{methodname} for {filetype} is already '
                                 f'registered for {classobj.__name__}')
            cls.registry[key] = wrapped_func
            return wrapped_func
        return inner_wrapper

    @classmethod
    def get_identifiers(cls, classobj):
        return [key for key in cls.registry
                if key[0] == classobj and key[1] == 'identify']

    @classmethod
    def _no_format_error(cls, classobj):
        msg = ('Format could not be identified based on the file name or '
               'contents, please provide a "format" argument.'
               f'\n{cls._get_format_table_str(classobj)}')
        raise IORegistryError(msg)

    @classmethod
    def identify_format(cls, filename, classobj, methodname):
        format = None
        identifiers = cls.get_identifiers(classobj)
        if identifiers:
            for identifier in identifiers:
                if cls.registry[identifier](methodname, filename):
                    format = identifier[2]
                    break  # finds the first valid filetype

        if format is None:
            cls._no_format_error(classobj)

        return format

    @classmethod
    def read(cls, filename, classobj, format=None, **kwargs):
        """
        Read in a regions file.
        """
        if format is None:
            format = cls.identify_format(filename, classobj, 'read')

        key = (classobj, 'read', format)
        try:
            return cls.registry[key](filename, **kwargs)
        except KeyError:
            msg = (f'No reader defined for format "{format}" and class '
                   f'"{classobj.__name__}".\n'
                   f'{cls._get_format_table_str(classobj)}')
            raise IORegistryError(msg) from None

    @classmethod
    def parse(cls, data, classobj, format=None, **kwargs):
        """
        Parse a regions string or table.
        """
        if format is None:
            cls._no_format_error(classobj)

        key = (classobj, 'parse', format)
        try:
            return cls.registry[key](data, **kwargs)
        except KeyError:
            msg = (f'No parser defined for format "{format}" and class '
                   f'"{classobj.__name__}".\n'
                   f'{cls._get_format_table_str(classobj)}')
            raise IORegistryError(msg) from None

    @classmethod
    def write(cls, regions, filename, classobj, format=None, **kwargs):
        """
        Write to a regions file.
        """
        if format is None:
            format = cls.identify_format(filename, classobj, 'write')

        key = (classobj, 'write', format)
        try:
            return cls.registry[key](regions, filename, **kwargs)
        except KeyError:
            msg = (f'No writer defined for format "{format}" and class '
                   f'"{classobj.__name__}".\n'
                   f'{cls._get_format_table_str(classobj)}')
            raise IORegistryError(msg) from None

    @classmethod
    def serialize(cls, regions, classobj, format=None, **kwargs):
        """
        Serialize to a regions string or table.
        """
        if format is None:
            cls._no_format_error(classobj)

        key = (classobj, 'serialize', format)
        try:
            return cls.registry[key](regions, **kwargs)
        except KeyError:
            msg = (f'No serializer defined for format "{format}" and class '
                   f'"{classobj.__name__}".\n'
                   f'{cls._get_format_table_str(classobj)}')
            raise IORegistryError(msg) from None

    @classmethod
    def get_formats(cls, classobj):
        """
        Get the registered I/O formats as a Table.
        """
        filetypes = list({key[2] for key in cls.registry
                          if key[0] == classobj})

        rows = [['Format', 'Parse', 'Serialize', 'Read', 'Write',
                 'Auto-identify']]
        for filetype in sorted(filetypes):
            keys = {key[1] for key in cls.registry
                    if key[0] == classobj and key[2] == filetype}
            row = [filetype]
            for methodname in rows[0][1:]:
                name = ('identify' if 'identify' in methodname
                        else methodname.lower())
                row.append('Yes' if name in keys else 'No')
            rows.append(row)

        if len(rows) == 1:
            return Table()

        cols = list(zip(*rows))
        tbl = Table()
        for col in cols:
            tbl[col[0]] = col[1:]

        return tbl

    @classmethod
    def _get_format_table_str(cls, classobj):
        lines = ['', f'The available formats for the {classobj.__name__} '
                 'class are:', '']
        tbl = cls.get_formats(classobj)
        lines.extend(tbl.pformat(max_lines=-1, max_width=80))
        return '\n'.join(lines)


def _update_docstring(classobj, methodname):
    """
    Update the docstring to include a table of all available registered
    formats and methods.
    """
    import re

    if methodname == 'identify':
        return

    lines = getattr(classobj, methodname).__doc__.splitlines()
    matches = [re.search(r'(\S)', line) for line in lines[1:]]
    left_indent = ' ' * min(match.start() for match in matches if match)
    new_lines = RegionsRegistry._get_format_table_str(classobj).splitlines()
    lines.extend([left_indent + line for line in new_lines])

    try:
        # classmethod
        getattr(classobj, methodname).__func__.__doc__ = '\n'.join(lines)
    except AttributeError:
        # instancemethod
        getattr(classobj, methodname).__doc__ = '\n'.join(lines)
    return



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
from numpy.testing import assert_allclose
import pytest

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.tests.helper import assert_quantity_allclose

from ...core import PixCoord, BoundingBox
from ...tests.helpers import make_simple_wcs
from ..._utils.examples import make_example_dataset
from ..polygon import PolygonPixelRegion, PolygonSkyRegion
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion
from .utils import HAS_MATPLOTLIB  # noqa


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    config = dict(crpix=(18, 9), cdelt=(-10, 10), shape=(18, 36))
    dataset = make_example_dataset(config=config)
    return dataset.wcs


class TestPolygonPixelRegion(BaseTestPixelRegion):

    reg = PolygonPixelRegion(PixCoord([1, 3, 1], [1, 1, 4]))
    sample_box = [0, 4, 0, 5]
    inside = [(2, 2)]
    outside = [(3, 2), (3, 3)]
    expected_area = 3
    expected_repr = ('<PolygonPixelRegion(vertices=PixCoord(x=[1 3 1], '
                     'y=[1 1 4]))>')
    expected_str = ('Region: PolygonPixelRegion\nvertices: '
                    'PixCoord(x=[1 3 1], y=[1 1 4])')

    # We will be using this polygon for basic tests:
    #
    # (3,0) *
    #       |\
    #       | \
    #       |  \
    # (0,0) *---* (2, 0)
    #

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.vertices.x, [1, 3, 1])
        assert_allclose(reg.vertices.y, [1, 1, 4])
        assert reg.visual == {}
        assert reg.meta == {}

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.vertices.x, self.reg.vertices.x)
        assert_allclose(reg_new.vertices.y, self.reg.vertices.y)

    def test_bounding_box(self):
        bbox = self.reg.bounding_box
        assert bbox == BoundingBox(ixmin=1, ixmax=4, iymin=1, iymax=5)

    def test_to_mask(self):
        # The true area of this polygon is 3
        # We only do very low-precision asserts below,
        # because results can be unstable with points
        # on the edge of the polygon.
        # Basically we check that mask.data is filled
        # with something useful at all.

        # Bounding box and output shape is independent of subpixels,
        # so we only assert on it once here, not in the other cases below
        mask = self.reg.to_mask(mode='center', subpixels=1)
        assert 2 <= np.sum(mask.data) <= 6
        assert mask.bbox == BoundingBox(ixmin=1, ixmax=4, iymin=1, iymax=5)
        assert mask.data.shape == (4, 3)

        # Test more cases for to_mask
        # This example is with the default: subpixels=5
        mask = self.reg.to_mask(mode='subpixels')
        assert 2 <= np.sum(mask.data) <= 6

        mask = self.reg.to_mask(mode='subpixels', subpixels=8)
        assert 2 <= np.sum(mask.data) <= 6

        mask = self.reg.to_mask(mode='subpixels', subpixels=9)
        assert 2 <= np.sum(mask.data) <= 6

        mask = self.reg.to_mask(mode='subpixels', subpixels=10)
        assert 2 <= np.sum(mask.data) <= 6

        with pytest.raises(NotImplementedError):
            self.reg.to_mask(mode='exact')

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        expected = [[1, 1], [3, 1], [1, 4], [1, 1]]
        assert_allclose(patch.xy, expected)

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(3, 1), -90 * u.deg)
        assert_allclose(reg.vertices.x, [3, 3, 6])
        assert_allclose(reg.vertices.y, [3, 1, 3])


class TestPolygonSkyRegion(BaseTestSkyRegion):

    reg = PolygonSkyRegion(SkyCoord([3, 4, 3] * u.deg, [3, 4, 4] * u.deg))

    expected_repr = ('<PolygonSkyRegion(vertices=<SkyCoord (ICRS): (ra, '
                     'dec) in deg\n    [(3., 3.), (4., 4.), (3., 4.)]>)>')
    expected_str = ('Region: PolygonSkyRegion\nvertices: <SkyCoord (ICRS):'
                    ' (ra, dec) in deg\n    [(3., 3.), (4., 4.), (3., 4.)]>')

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.vertices.ra.deg, [3, 4, 3])
        assert reg.visual == {}
        assert reg.meta == {}

    def test_transformation(self, wcs):

        pixpoly = self.reg.to_pixel(wcs)

        assert_allclose(pixpoly.vertices.x, [11.187992, 10.976332, 11.024032],
                        atol=1e-5)
        assert_allclose(pixpoly.vertices.y, [1.999486, 2.039001, 2.077076],
                        atol=1e-5)

        poly = pixpoly.to_sky(wcs)

        # TODO: we should probably assert something about frame
        # attributes, or generally some better way to check if two
        # SkyCoord are the same? For now, we use the folloing line to
        # transform back to ICRS (`poly` is in Galactic, same as WCS)
        poly = PolygonSkyRegion(
            vertices=poly.vertices.transform_to(self.reg.vertices))
        assert_quantity_allclose(poly.vertices.data.lon,
                                 self.reg.vertices.data.lon, atol=1e-3 * u.deg)
        assert_quantity_allclose(poly.vertices.data.lat,
                                 self.reg.vertices.data.lat, atol=1e-3 * u.deg)

    def test_contains(self, wcs):
        position = SkyCoord([1, 3.25] * u.deg, [2, 3.75] * u.deg)
        # 1,2 is outside, 3.25,3.75 should be inside the triangle...
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, True], dtype='bool'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import warnings

from astropy.io import fits

from ...core import Region, Regions
from ...core.registry import RegionsRegistry

__all__ = []


@RegionsRegistry.register(Region, 'identify', 'fits')
@RegionsRegistry.register(Regions, 'identify', 'fits')
def is_fits(methodname, filepath):
    """
    Identify a FITS region file.

    Parameters
    ----------
    methodname : {'read', 'write'}
        The method name called that needs auto-identification.

    filepath : str
        The path to the file.

    Returns
    -------
    result : bool
        Returns `True` if the given file is a FITS region file.
    """
    all_exten = ('.fits', '.fit', '.fts', '.fits.gz', '.fit.gz', '.fts.gz')
    exten = {'read': all_exten, 'write': all_exten[0:3]}

    if methodname == 'write':
        return filepath.lower().endswith(exten[methodname])

    elif methodname == 'read':
        if (isinstance(filepath, str)
                and filepath.lower().endswith(exten[methodname])):
            return True
        else:
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    with fits.open(filepath):
                        return True
            except IOError:
                return False

    else:
        return False



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.table import vstack as table_vstack
from astropy.utils import lazyproperty
from astropy.wcs import WCS
import numpy as np

from ..core.pixcoord import PixCoord

__all__ = ['make_example_dataset']


def make_example_dataset(data='simulated', config=None):
    """
    Make an example dataset.

    This is a factory function for ``ExampleDataset`` objects.

    The following config options are available (default values shown):

    * ``crval = 0, 0``
    * ``crpix = 180, 90``
    * ``cdelt = -1, 1``
    * ``shape = 180, 360``
    * ``ctype = 'GLON-AIT', 'GLAT-AIT'``

    Parameters
    ----------
    data : {'simulated', 'fermi'}
        The dataset to use.

    config : dict or None
        Configuration options.

    Returns
    -------
    dataset : ``ExampleDataset``
        An example dataset object.

    Examples
    --------
    Make an example dataset:

    >>> from regions import make_example_dataset
    >>> config = dict(crpix=(18, 9), cdelt=(-10, 10), shape=(18, 36))
    >>> dataset = make_example_dataset(data='simulated', config=config)

    Access properties of the ``dataset`` object:

    >>> dataset.source_table   # doctest: +IGNORE_OUTPUT
    >>> dataset.event_table   # doctest: +IGNORE_OUTPUT
    >>> dataset.wcs   # doctest: +IGNORE_OUTPUT
    >>> dataset.image   # doctest: +IGNORE_OUTPUT
    >>> dataset.hdu_list   # doctest: +IGNORE_OUTPUT
    """
    if data == 'simulated':
        return ExampleDatasetSimulated(config=config)
    elif data == 'fermi':
        return ExampleDatasetFermi(config=config)
    else:
        raise ValueError(f'Invalid selection data: {data}')


class ExampleDataset:
    """
    Base class for an example dataset.
    """

    def __init__(self, config=None):
        config_default = dict()
        # These parameters are FITS WCS order: (x, y)
        config_default['crval'] = 0, 0
        config_default['crpix'] = 180, 90
        config_default['cdelt'] = -1, 1

        # This parameter is numpy order: (y, x)
        config_default['shape'] = 180, 360
        config_default['ctype'] = 'GLON-AIT', 'GLAT-AIT'

        # Callers can update the default config parameters they want.
        if config:
            config_default.update(config)

        self.config = config_default

    @lazyproperty
    def wcs(self):
        """World coordinate system (`~astropy.wcs.WCS`)."""
        wcs = WCS(naxis=2)
        wcs.wcs.crval = self.config['crval']
        wcs.wcs.crpix = self.config['crpix']
        wcs.wcs.cdelt = self.config['cdelt']
        wcs.wcs.ctype = self.config['ctype']

        return wcs

    @lazyproperty
    def image(self):
        """Counts image (`~astropy.io.fits.ImageHDU`)."""
        events = self.event_table
        skycoord = SkyCoord(events['GLON'], events['GLAT'], unit='deg',
                            frame='galactic')
        pixcoord = PixCoord.from_sky(skycoord=skycoord, wcs=self.wcs)

        shape = self.config['shape']
        bins = [np.arange(shape[0] + 1), np.arange(shape[1] + 1)]
        sample = np.vstack((pixcoord.y, pixcoord.x)).T
        data, _ = np.histogramdd(sample=sample, bins=bins)
        data = data.astype('float32')

        header = self.wcs.to_header()
        return fits.ImageHDU(data=data, header=header, name='image')

    @lazyproperty
    def hdu_list(self):
        """
        HDU list (`~astropy.io.fits.HDUList`).

        Different pieces collected together in a HDU list.

        This method makes it easy to write the example dataset to a FITS
        file with multiple HDUs.
        """
        hdu_list = fits.HDUList()

        hdu = _table_to_bintable(self.source_table)
        hdu.name = 'sources'
        hdu_list.append(hdu)

        hdu = _table_to_bintable(self.event_table)
        hdu.name = 'events'
        hdu_list.append(hdu)

        hdu = self.image
        hdu.name = 'image'
        hdu_list.append(hdu)

        return hdu_list


class ExampleDatasetSimulated(ExampleDataset):
    """
    Example simulated dataset.

    Similar to `ExampleDatasetFermi`, but simulated, not requiring any
    data files.
    """

    @lazyproperty
    def source_table(self):
        """
        Source table  (`~astropy.table.Table`).

        Columns: GLON, GLAT, COUNTS
        """
        table = Table()
        table['GLON'] = np.array([0, 45, 45], dtype='float32')
        table['GLAT'] = np.array([0, 0, 45], dtype='float32')
        table['COUNTS'] = np.array([100, 100, 100], dtype='int32')
        return table

    @lazyproperty
    def event_table(self):
        """
        Event table (`~astropy.table.Table`).

        Columns: GLON, GLAT, SOURCE_IDX
        """
        # Create event list table for each source
        tables = []
        for source in self.source_table:
            lon = source['GLON'] * np.ones(source['COUNTS'])
            lat = source['GLAT'] * np.ones(source['COUNTS'])

            # TODO: scatter positions assuming Gaussian PSF on the sky
            # using SkyOffsetFrame.
            # coord = SkyCoord(lon, lat, unit='deg', frame='galactic')

            table = Table()
            table['GLON'] = lon
            table['GLAT'] = lat
            table['SOURCE_IDX'] = source.index

            tables.append(table)

        # Stack all tables together
        table = table_vstack(tables)

        return table


class ExampleDatasetFermi(ExampleDataset):
    """
    Example real dataset using Fermi-LAT 2FHL source catalog and event
    list.

    When saving the HDU list to a FITS file, the file size is 748K, with
    the 65k EVENTS taking up most of the space.
    """

    @lazyproperty
    def source_table(self):
        """
        Source table  (`~astropy.table.Table`).

        Columns: GLON, GLAT, COUNTS
        """
        url = ('https://github.com/gammapy/gammapy-extra/raw/master/datasets/'
               'fermi_2fhl/gll_psch_v08.fit.gz')
        table = Table.read(url, hdu='2FHL Source Catalog')
        table.rename_column('Npred', 'COUNTS')
        table.keep_columns(['GLON', 'GLAT', 'COUNTS'])
        table.meta.clear()
        return table

    @lazyproperty
    def event_table(self):
        """
        Event table (`~astropy.table.Table`).

        Columns: GLON, GLAT
        """
        url = ('https://github.com/gammapy/gammapy-extra/raw/master/datasets/'
               'fermi_2fhl/2fhl_events.fits.gz')
        table = Table.read(url, hdu='EVENTS')
        table.rename_column('L', 'GLON')
        table.rename_column('B', 'GLAT')
        table.keep_columns(['GLON', 'GLAT'])
        table.meta.clear()
        return table


def _table_to_bintable(table):
    """
    Convert a `~astropy.table.Table` to a `astropy.io.fits.BinTable`.
    """
    data = table.as_array()
    header = fits.Header()
    header.update(table.meta)
    name = table.meta.pop('name', None)
    return fits.BinTableHDU(data, header, name=name)


if __name__ == '__main__':
    dataset = make_example_dataset(data='simulated')
    dataset.hdu_list.writeto('example-dataset-simulated.fits', overwrite=True)

    dataset = make_example_dataset(data='fermi')
    dataset.hdu_list.writeto('example-dataset-fermi.fits', overwrite=True)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the pixcoord module.
"""

import astropy.units as u
import numpy as np
from numpy.testing import assert_equal, assert_allclose
import pytest

from ..._utils.examples import make_example_dataset
from ..pixcoord import PixCoord


@pytest.fixture(scope='session', name='wcs')
def fixture_wcs():
    config = dict(crpix=(18, 9), cdelt=(-10, 10), shape=(18, 36))
    dataset = make_example_dataset(config=config)
    return dataset.wcs


def test_pixcoord_copy_scalar():
    pc1 = PixCoord(x=1, y=2)
    pc2 = pc1.copy()
    pc1.x = 99
    pc1.y = 99
    assert pc2.xy == (1, 2)


def test_pixcoord_copy_array():
    pc1 = PixCoord(x=[1, 2], y=[3, 4])
    pc2 = pc1.copy()
    pc1.x[0] = 99
    pc1.y[0] = 99
    assert_equal(pc2.x, [1, 2])
    assert_equal(pc2.y, [3, 4])


def test_pixcoord_basic_dimension():
    with pytest.raises(ValueError):
        PixCoord(np.array([1, 2]), [3, 4, 5, 6])


def test_pixcoord_basics_scalar():
    pc0 = PixCoord(x=1, y=2)
    pc1 = PixCoord(x=np.array(1), y=2)
    pc2 = PixCoord(x=np.array(1), y=np.array(2))

    assert pc0 == pc1
    assert pc1 == pc2

    assert pc0.x == 1
    assert pc0.y == 2

    assert pc0.isscalar
    assert pc1.isscalar
    assert pc2.isscalar

    assert str(pc0) == 'PixCoord(x=1, y=2)'
    assert repr(pc0) == 'PixCoord(x=1, y=2)'

    with pytest.raises(TypeError):
        _ = len(pc0)

    with pytest.raises(IndexError):
        _ = pc0[0]


def test_pixcoord_basics_array_1d():
    pc0 = PixCoord(x=[1, 2, 3], y=[11, 22, 33])
    pc1 = PixCoord(x=1, y=[1, 2, 3])

    assert_equal(pc0.x, [1, 2, 3])
    assert_equal(pc0.y, [11, 22, 33])

    assert_equal(pc1.x, [1, 1, 1])
    assert_equal(pc1.y, [1, 2, 3])

    assert not pc0.isscalar
    assert not pc1.isscalar

    assert str(pc0) == 'PixCoord(x=[1 2 3], y=[11 22 33])'
    assert str(pc1) == 'PixCoord(x=[1 1 1], y=[1 2 3])'

    assert repr(pc0) == 'PixCoord(x=[1 2 3], y=[11 22 33])'
    assert repr(pc1) == 'PixCoord(x=[1 1 1], y=[1 2 3])'

    assert len(pc0) == 3
    assert len(pc1) == 3

    pc2 = list(iter(pc0))[-1]
    assert pc2.x == 3
    assert pc2.y == 33

    pc3 = pc0[-1]
    assert pc3.isscalar
    assert pc3.x == 3
    assert pc3.y == 33

    pc4 = pc0[1:]
    assert len(pc4) == 2
    assert_equal(pc4.x, [2, 3])
    assert_equal(pc4.y, [22, 33])


def test_pixcoord_basics_array_2d():
    pc0 = PixCoord([[1, 2, 3], [4, 5, 6]], [[11, 12, 13], [14, 15, 16]])
    assert_equal(pc0.x, [[1, 2, 3], [4, 5, 6]])
    assert_equal(pc0.y, [[11, 12, 13], [14, 15, 16]])
    assert pc0.isscalar is False

    expected = ('PixCoord(x=[[1 2 3]\n [4 5 6]], '
                'y=[[11 12 13]\n [14 15 16]])')
    assert str(pc0) == expected
    assert repr(pc0) == expected
    assert len(pc0) == 2

    assert_equal(pc0[0].x, pc0.x[0])
    assert_equal(pc0[0].y, pc0.y[0])

    pc1 = list(iter(pc0))
    assert_equal(pc1[0].x, pc0.x[0])
    assert_equal(pc1[0].y, pc0.y[0])


def test_pixcoord_to_sky_scalar(wcs):
    pc1 = PixCoord(x=18, y=9)

    sc = pc1.to_sky(wcs=wcs)
    assert sc.name == 'galactic'
    assert_allclose(sc.data.lon.deg, 349.88093995435634)
    assert_allclose(sc.data.lat.deg, 10.003028030623508)

    pc2 = PixCoord.from_sky(skycoord=sc, wcs=wcs)
    assert isinstance(pc2.x, float)
    assert pc2.isscalar

    assert pc1 == pc2


def test_pixcoord_to_sky_array_1d(wcs):
    pc1 = PixCoord(x=[17, 18], y=[8, 9])

    sc = pc1.to_sky(wcs=wcs)
    assert sc.name == 'galactic'
    assert_allclose(sc.data.lon.deg, [0, 349.88094])
    assert_allclose(sc.data.lat.deg, [0, 10.003028])

    pc2 = PixCoord.from_sky(skycoord=sc, wcs=wcs)
    assert isinstance(pc2.x, np.ndarray)
    assert pc2.x.shape == (2,)
    assert not pc2.isscalar

    assert pc1 == pc2


def test_pixcoord_to_sky_array_2d(wcs):
    pc0 = PixCoord(x=[[17, 18]], y=[[8, 9]])
    pc1 = PixCoord(x=[[17, 17, 17], [18, 18, 18]], y=[[8, 8, 8], [9, 9, 9]])

    sc0 = pc0.to_sky(wcs=wcs)
    assert sc0.name == 'galactic'

    sc1 = pc1.to_sky(wcs=wcs)
    assert sc1.name == 'galactic'
    assert_allclose(sc1.data.lon.deg, [[0, 0, 0],
                                       [349.88094, 349.88094, 349.88094]])
    assert_allclose(sc1.data.lat.deg, [[0, 0, 0],
                                       [10.003028, 10.003028, 10.003028]])
    assert_allclose(sc0.data.lon.deg, [[0, 349.88094]])
    assert_allclose(sc0.data.lat.deg, [[0, 10.003028]])

    pc2 = PixCoord.from_sky(skycoord=sc0, wcs=wcs)
    assert isinstance(pc2.x, np.ndarray)
    assert pc2.x.shape == (1, 2)
    assert not pc2.isscalar

    assert pc0 == pc2


def test_pixcoord_separation_scalar():
    pc1 = PixCoord(x=1, y=2)
    pc2 = PixCoord(x=4, y=6)
    sep = pc1.separation(pc2)
    assert_allclose(sep, 5)


def test_pixcoord_separation_array_1d():
    pc1 = PixCoord(x=[1, 1], y=[2, 2])
    pc2 = PixCoord(x=[4, 4], y=[6, 6])
    sep = pc1.separation(pc2)
    assert_allclose(sep, [5, 5])


def test_pixcoord_separation_array_2d():
    pc1 = PixCoord(x=[[1, 1]], y=[[2, 2]])
    pc2 = PixCoord(x=[[4, 4]], y=[[6, 6]])
    sep = pc1.separation(pc2)
    assert isinstance(sep, np.ndarray)
    assert sep.shape == (1, 2)
    assert_allclose(sep, [[5, 5]])


def test_equality():
    arr = np.array([1, 2])
    pc1 = PixCoord(arr[0], arr[1])
    pc2 = PixCoord(arr[0] + 0.0000001, arr[1])

    assert not pc1 == arr
    assert pc1 == PixCoord(arr[0], arr[1])
    assert pc1 == pc2

    pc3 = PixCoord([[1, 2, 3], [4, 5, 6]], [[11, 12, 13], [14, 15, 16]])
    pc4 = PixCoord([[1, 2, 3], [4, 5, 6]],
                   [[11.0000002, 12, 13], [14, 15, 16]])

    assert pc3 == pc4
    assert pc3 == PixCoord([[1, 2, 3], [4, 5, 6]],
                           [[11, 12, 13], [14, 15, 16]])


def test_pixcoord_xy():
    arr = np.array([1, 2])
    pc = PixCoord(arr[0], arr[1])
    assert pc.xy[0] == pc.x
    assert pc.xy[1] == pc.y


def test_pixcoord_rotate_scalar():
    point = PixCoord(3, 4)
    center = PixCoord(2, 3)
    point = point.rotate(center, 90 * u.deg)
    assert_allclose(point.xy, (1, 4))


def test_pixcoord_rotate_array():
    point = PixCoord([3, 3], [4, 4])
    center = PixCoord(2, 3)
    point = point.rotate(center, 90 * u.deg)
    assert_allclose(point.xy, ([1, 1], [4, 4]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from numpy.testing import assert_allclose
import numpy as np
import pytest

from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
from astropy.wcs import WCS

from ...core import PixCoord
from ...tests.helpers import make_simple_wcs
from ..line import LinePixelRegion, LineSkyRegion
from .test_common import BaseTestPixelRegion, BaseTestSkyRegion
from .utils import HAS_MATPLOTLIB  # noqa


@pytest.fixture(scope='session', name='wcs')
def wcs_fixture():
    filename = get_pkg_data_filename('data/example_header.fits')
    header = fits.getheader(filename)
    return WCS(header)


class TestLinePixelRegion(BaseTestPixelRegion):

    reg = LinePixelRegion(PixCoord(3, 4), PixCoord(4, 4))
    sample_box = [-2, 8, -1, 9]
    inside = []
    outside = [(3.1, 4.2), (5, 4)]
    expected_area = 0
    expected_repr = ('<LinePixelRegion(start=PixCoord(x=3, y=4), '
                     'end=PixCoord(x=4, y=4))>')
    expected_str = ('Region: LinePixelRegion\nstart: PixCoord(x=3, y=4)\n'
                    'end: PixCoord(x=4, y=4)')

    def test_copy(self):
        reg = self.reg.copy()
        assert reg.start.xy == (3, 4)
        assert reg.end.xy == (4, 4)
        assert reg.visual == {}
        assert reg.meta == {}

    def test_pix_sky_roundtrip(self):
        wcs = make_simple_wcs(SkyCoord(2 * u.deg, 3 * u.deg), 0.1 * u.deg, 20)
        reg_new = self.reg.to_sky(wcs).to_pixel(wcs)
        assert_allclose(reg_new.start.x, self.reg.start.x)
        assert_allclose(reg_new.start.y, self.reg.start.y)
        assert_allclose(reg_new.end.x, self.reg.end.x)
        assert_allclose(reg_new.end.y, self.reg.end.y)

    @pytest.mark.skipif('not HAS_MATPLOTLIB')
    def test_as_artist(self):
        patch = self.reg.as_artist()
        assert 'Arrow' in str(patch)

    def test_rotate(self):
        reg = self.reg.rotate(PixCoord(2, 3), 90 * u.deg)
        assert_allclose(reg.start.xy, (1, 4))
        assert_allclose(reg.end.xy, (1, 5))


class TestLineSkyRegion(BaseTestSkyRegion):

    start = SkyCoord(3 * u.deg, 4 * u.deg, frame='galactic')
    end = SkyCoord(3 * u.deg, 5 * u.deg, frame='galactic')
    reg = LineSkyRegion(start, end)

    expected_repr = ('<LineSkyRegion(start=<SkyCoord (Galactic): (l, b) '
                     'in deg\n    (3., 4.)>, end=<SkyCoord (Galactic): '
                     '(l, b) in deg\n    (3., 5.)>)>')
    expected_str = ('Region: LineSkyRegion\nstart: <SkyCoord (Galactic): '
                    '(l, b) in deg\n    (3., 4.)>\nend: <SkyCoord '
                    '(Galactic): (l, b) in deg\n    (3., 5.)>')

    def test_copy(self):
        reg = self.reg.copy()
        assert_allclose(reg.start.b.deg, 4)
        assert_allclose(reg.end.b.deg, 5)
        assert reg.visual == {}
        assert reg.meta == {}

    def test_transformation(self, wcs):
        pixline = self.reg.to_pixel(wcs)

        assert_allclose(pixline.start.x, -50.5)
        assert_allclose(pixline.start.y, 299.5)
        assert_allclose(pixline.end.x, -50.5)
        assert_allclose(pixline.end.y, 349.5)

        skyline = pixline.to_sky(wcs)

        assert_quantity_allclose(skyline.start.data.lon,
                                 self.reg.start.data.lon)
        assert_quantity_allclose(skyline.start.data.lat,
                                 self.reg.start.data.lat)
        assert_quantity_allclose(skyline.end.data.lon, self.reg.end.data.lon)
        assert_quantity_allclose(skyline.end.data.lat, self.reg.end.data.lat)

    def test_contains(self, wcs):
        position = SkyCoord([1, 2] * u.deg, [3, 4] * u.deg)
        # lines do not contain things
        assert all(self.reg.contains(position, wcs)
                   == np.array([False, False], dtype='bool'))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.utils.data import get_pkg_data_filename, get_pkg_data_fileobj
import pytest

from ...core import Regions


parametrize_filenames = pytest.mark.parametrize(
    'filename', ['../crtf/tests/data/CRTFgeneral.crtf',
                 '../ds9/tests/data/ds9.color.reg',
                 '../fits/tests/data/fits_region.fits'])


@parametrize_filenames
def test_read_filename(filename):
    filename = get_pkg_data_filename(filename)
    result = Regions.read(filename)
    assert isinstance(result, Regions)


@parametrize_filenames
@pytest.mark.parametrize('encoding',
                         ['binary',
                          pytest.param(None, marks=pytest.mark.xfail)])
def test_read_fileobj(filename, encoding):
    with get_pkg_data_fileobj(filename, encoding=encoding) as fileobj:
        result = Regions.read(fileobj)
    assert isinstance(result, Regions)


@pytest.mark.parametrize('file_type', ['crtf', 'ds9'])
def test_write_format(file_type, tmpdir):
    infilename = get_pkg_data_filename('../crtf/tests/data/CRTFgeneral.crtf')
    outfilename = str(tmpdir / 'region')
    shapelist = Regions.read(infilename)
    shapelist.write(outfilename, format=file_type)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the crtf subpackage.
"""

from astropy.coordinates import Angle, SkyCoord
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
import pytest

from ....shapes.circle import CircleSkyRegion
from ....shapes.ellipse import EllipseSkyRegion
from ....core import Regions
from ..core import CRTFRegionParserError
from ..read import _CRTFParser


implemented_region_types = ('ellipse', 'circle', 'rectangle', 'poly', 'point',
                            'text', 'symbol')


def test_global_parser():
    """
    Checks that the global_parser does what's expected.
    """
    global_test_str = ('global coord=B1950_VLA, frame=BARY, corr=[I, Q], '
                       'color=blue')
    global_parser = _CRTFParser(global_test_str)
    expected = {'coord': 'B1950_VLA', 'frame': 'BARY',
                'corr': ['I', 'Q'], 'color': 'blue'}
    assert dict(global_parser.global_meta) == expected


def test_valid_crtf_line():
    """
    Checks whether a the line is valid CRTF format.
    """
    line_str = 'coord=B1950_VLA, frame=BARY, corr=[I, Q], color=blue'

    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(line_str, format='crtf')

    assert 'Not a valid CRTF line:' in str(excinfo.value)


def test_valid_region_type():
    """
    Checks whether the region type is valid in CRTF format
    """
    reg_str = 'hyperbola[[18h12m24s, -23d11m00s], 2.3arcsec]'

    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str, format='crtf')

    assert 'Not a valid CRTF Region type: "hyperbola"' in str(excinfo.value)


def test_valid_global_meta_key():
    """
    Checks whether the global key is valid or not.
    """
    global_test_str = ('global label=B1950_VLA, frame=BARY, corr=[I, Q], '
                       'color=blue')
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(global_test_str, format='crtf')
    assert '"label" is not a valid global meta key' in str(excinfo.value)


def test_valid_meta_key():
    """
    Checks whether the key is valid or not.
    """
    meta_test_str = ('annulus[[17h51m03.2s, -45d17m50s], [0.10deg, 4.12deg]], '
                     'hello="My label here"')
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(meta_test_str, format='crtf')
    assert '"hello" is not a valid meta key' in str(excinfo.value)


def test_valid_region_syntax():
    """
    Checks whether the region has valid parameters
    """
    reg_str1 = 'circle[[18h12m24s, -23d11m00s], [2.3arcsec,4.5arcsec]'
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str1, format='crtf')

    estr = ("Not in proper format: ('2.3arcsec', '4.5arcsec') should be "
            "a single length")
    assert estr in str(excinfo.value)

    reg_str2 = ('symbol[[32.1423deg, 12.1412deg], 12deg], linewidth=2, '
                'coord=J2000, symsize=2')
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str2, format='crtf')
    estr = 'Not in proper format: "12deg" should be a symbol'
    assert estr in str(excinfo.value)

    reg_str3 = 'circle[[18h12m24s, -23d11m00s]'
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str3, format='crtf')
    estr = ('Does not contain expected number of parameters for the region '
            '"circle"')
    assert estr in str(excinfo.value)

    reg_str4 = 'poly[[1, 2], [4, 5]]'
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str4, format='crtf')
    assert 'polygon should have >= 3 coordinates' in str(excinfo.value)

    reg_str6 = 'rotbox[[12h01m34.1s, 12d23m33s], [3arcmin,], 12deg]'
    with pytest.raises(CRTFRegionParserError) as excinfo:
        Regions.parse(reg_str6, format='crtf')
    assert "('3arcmin', '') should be a pair of lengths" in str(excinfo.value)


def test_issue_312_regression():
    """
    Make sure there is no trailing comma when writing a CRTF string with
    no metadata.
    """
    reg = EllipseSkyRegion(center=SkyCoord(279.174990 * u.deg,
                                           -21.257123 * u.deg, frame='fk5'),
                           width=0.001571 * u.deg, height=0.001973 * u.deg,
                           angle=111.273322 * u.deg)
    crtfstr = reg.serialize(format='crtf', coordsys='fk5', fmt='.6f',
                            radunit='deg')
    assert crtfstr.strip()[-1] != ','


@pytest.mark.parametrize('filename', ['data/CRTFgeneral.crtf',
                                      'data/CRTFgeneraloutput.crtf'])
def test_file_crtf(filename):
    filename = get_pkg_data_filename(filename)
    regs = Regions.read(filename, errors='warn', format='crtf')
    actual_output = regs.serialize(format='crtf', coordsys='fk4',
                                   fmt='.3f').strip()

    with open(get_pkg_data_filename('data/CRTFgeneraloutput.crtf')) as f:
        ref_output = f.read().strip()

    # since metadata is not required to preserve order, we have to do a more
    # complex comparison
    desired_lines = [set(line.split(",")) for line in ref_output.split("\n")]
    actual_lines = [set(line.split(",")) for line in actual_output.split("\n")]
    for split_line in actual_lines:
        assert split_line in desired_lines

    for split_line in desired_lines:
        assert split_line in actual_lines


def test_crtf_header():
    """
    Regression test for issue #239.
    """
    crtf_str = ('#CRTFv0 CASA Region Text Format version 0\n'
                'circle[[42deg, 43deg], 3deg], coord=J2000, color=green')
    reg = Regions.parse(crtf_str, format='crtf')[0]
    assert isinstance(reg, CircleSkyRegion)
    assert reg.center.ra.value == 42.0
    assert reg.center.ra.unit == 'deg'
    assert reg.center.dec.value == 43.0
    assert reg.center.dec.unit == 'deg'
    assert reg.radius.value == 3.0
    assert reg.radius.unit == 'deg'


def test_space_after_regname():
    """
    Regression test for #271: space is allowed
    """
    reg_str = 'circle [[42deg, 43deg], 3deg], coord=J2000, color=green'
    reg = Regions.parse(reg_str, format='crtf')[0]
    assert isinstance(reg, CircleSkyRegion)


def test_no_comma_after_region():
    reg_str = 'circle [[42deg, 43deg], 3deg] coord=J2000, color=green'
    reg = Regions.parse(reg_str, format='crtf')[0]
    assert isinstance(reg, CircleSkyRegion)
    assert reg.center.ra.value == 42.0
    assert reg.center.ra.unit == 'deg'
    assert reg.center.dec.value == 43.0
    assert reg.center.dec.unit == 'deg'
    assert reg.radius.value == 3.0
    assert reg.radius.unit == 'deg'


def test_casa_file_crtf():
    filename = get_pkg_data_filename('data/CRTF_CARTA.crtf')
    regions = Regions.read(filename, format='crtf')
    assert len(regions) == 2


def test_angle_serialization():
    """
    Regression test for issue #223 to ensure Angle arcsec inputs are
    correctly converted to degrees.
    """
    reg = Regions([CircleSkyRegion(SkyCoord(10,20, unit='deg'),
                                   Angle(1, 'arcsec'))])
    regstr = reg.serialize(format='crtf')
    expected = ('#CRTFv0\nglobal coord=J2000\ncircle[[10.000009deg, '
                '20.000002deg], 0.000278deg]\n')
    assert regstr == expected



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.io import fits
from astropy.utils.decorators import deprecated

from ...core import Region, Regions
from ...core.registry import RegionsRegistry
from ..core import _to_shape_list, SkyRegion

__all__ = ['write_fits', 'fits_region_objects_to_table']


@RegionsRegistry.register(Region, 'serialize', 'fits')
@RegionsRegistry.register(Regions, 'serialize', 'fits')
def _serialize_fits(regions):
    for region in regions:
        if isinstance(region, SkyRegion):
            raise TypeError('Every region must be a pixel region')

    shape_list = _to_shape_list(regions, coordinate_system='image')
    return shape_list.to_fits()


@deprecated('0.5', alternative='`regions.Regions.serialize`')
def fits_region_objects_to_table(regions):
    """
    Convert a list of `~regions.Region` objects to a FITS region table.

    Parameters
    ----------
    regions : list
        A list of `regions.Region` objects.

    Returns
    -------
    region_string : `~astropy.table.Table`
        A FITS region table.
    """
    return _serialize_fits(regions)


@RegionsRegistry.register(Region, 'write', 'fits')
@RegionsRegistry.register(Regions, 'write', 'fits')
def _write_fits(regions, filename, header=None, overwrite=False):
    """
    Convert a list of `~regions.Region` to a FITS region table and write
    to a file.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    filename : str
        The filename in which the table is to be written.

    header : `~astropy.io.fits.Header`, optional
        The FITS header.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    output = _serialize_fits(regions)
    bin_table = fits.BinTableHDU(data=output, header=header)
    bin_table.writeto(filename, overwrite=overwrite)


@deprecated('0.5', alternative='`regions.Regions.write`')
def write_fits_region(filename, regions, header=None, overwrite=False):
    """
    Convert a list of `~regions.Region` to a FITS region table and write
    to a file.

    Parameters
    ----------
    filename : str
        The filename in which the table is to be written.

    regions : list
        A list of `~regions.Region` objects.

    header : `~astropy.io.fits.Header`, optional
        The FITS header.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    _write_fits(regions, filename, header=header, overwrite=overwrite)


@deprecated('0.5', alternative='`regions.Regions.write`')
def write_fits(filename, regions, header=None, overwrite=False):
    """
    Convert a list of `~regions.Region` to a FITS region table and write
    to a file.

    Parameters
    ----------
    filename : str
        The filename in which the table is to be written.

    regions : list
        A list of `~regions.Region` objects.

    header : `~astropy.io.fits.Header`, optional
        The FITS header.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    _write_fits(regions, filename, header=header, overwrite=overwrite)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the fits subpackage.
"""

from astropy.coordinates import SkyCoord
from astropy.io import fits
import astropy.units as u
from astropy.utils.data import get_pkg_data_filename
from astropy.table import Table
from astropy.wcs import WCS
from numpy.testing import assert_allclose
import pytest

from ....core import Regions
from ....shapes import CircleSkyRegion
from ...core import _to_shape_list
from ..core import FITSRegionParserError
from ..read import _FITSRegionParser

implemented_region_types = ('ellipse', 'circle', 'box', 'polygon', 'point',
                            'annulus', 'elliptannulus')


@pytest.mark.parametrize('filename', ['data/fits_region.fits'])
def test_file_fits(filename):
    filename = get_pkg_data_filename(filename)
    table = Table.read(filename)
    shapes = _FITSRegionParser(table, 'warn').shapes

    assert shapes[0].region_type == 'circle'
    assert shapes[1].region_type == 'rectangle'
    assert shapes[2].region_type == 'ellipse'
    assert shapes[3].region_type == 'rectangle'
    assert shapes[4].region_type == 'circleannulus'
    assert shapes[5].region_type == 'point'
    assert shapes[6].region_type == 'point'
    assert shapes[7].region_type == 'polygon'
    assert shapes[8].region_type == 'rectangle'

    for x in range(8):
        assert_allclose(shapes[x].coord[:2],
                        [table['X'][x][0], table['Y'][x][0]])

    assert_allclose(shapes[7].coord, [1, 5, 2, 6, 3, 7, 4, 8])

    assert_allclose(shapes[0].coord[2:], [table['R'][0][0]])

    for x in range(1, 4):
        assert_allclose([val.value for val in shapes[x].coord[2:]],
                        list(table['R'][x][:2]) + [table['ROTANG'][x]])

    regs = Regions(shapes.to_regions())
    table_output = regs.serialize(format='fits')
    shape_output = _FITSRegionParser(table_output).shapes

    for i, shape in enumerate(shapes):
        assert shape.region_type == shape_output[i].region_type
        assert shape.coord == shape_output[i].coord
        assert shape.meta == shape_output[i].meta

    # Reading the regions directly from file and converting to sky
    # regions.
    regs_sky = Regions.read(filename, format='fits')
    with fits.open(filename) as hdulist:
        header = hdulist[1].header
        wcs = WCS(header, keysel=['image', 'binary', 'pixel'])
        regs_pix = [reg.to_pixel(wcs) for reg in regs_sky]
        shapes_roundtrip = _to_shape_list(regs_pix, 'image')

    for i, shape in enumerate(shapes):
        assert shape.region_type == shapes_roundtrip[i].region_type
        assert_allclose(shape.coord[:-1], shapes_roundtrip[i].coord[:-1])


def test_only_pixel_regions():
    reg_sky = CircleSkyRegion(SkyCoord(1, 2, unit='deg'), 5 * u.deg)
    reg = Regions([reg_sky])

    with pytest.raises(TypeError) as excinfo:
        reg.serialize(format='fits')

    estr = 'Every region must be a pixel region'
    assert estr in str(excinfo.value)


def test_valid_columns():
    t = Table([[1, 2, 3]], names=('a'))

    with pytest.raises(FITSRegionParserError) as excinfo:
        Regions.parse(t, format='fits')

    estr = 'This table has an invalid column name: "a"'
    assert estr in str(excinfo.value)


def test_valid_row():
    x = [1]
    y = [2]
    shapes = ['CIRCLE']
    t = Table([x, y, shapes], names=('X', 'Y', 'SHAPE'))
    t['X'].unit = 'pix'
    t['Y'].unit = 'pix'

    with pytest.raises(FITSRegionParserError) as excinfo:
        Regions.parse(t, format='fits')

    estr = 'The column "R" is missing in the table'
    assert estr in str(excinfo.value)

    t[0]['SHAPE'] = 'PONT'

    with pytest.raises(FITSRegionParserError) as excinfo:
        Regions.parse(t, format='fits')

    estr = '"PONT" is not a valid FITS Region type'
    assert estr in str(excinfo.value)

    t['ROTANG'] = [[20, 30]]
    t['ROTANG'].unit = 'deg'
    t[0]['SHAPE'] = 'PIE'

    with pytest.raises(FITSRegionParserError) as excinfo:
        Regions.parse(t, format='fits')

    estr = '"PIE" is currently not supported'
    assert estr in str(excinfo.value)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines polygon regions in both pixel and sky coordinates.
"""

from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel
import numpy as np

from ..core.attributes import OneDPix, OneDSky
from ..core.bounding_box import BoundingBox
from ..core.core import PixelRegion, SkyRegion
from ..core.mask import RegionMask
from ..core.metadata import RegionMeta, RegionVisual
from ..core.pixcoord import PixCoord
from .._geometry import polygonal_overlap_grid
from .._geometry.pnpoly import points_in_polygon

__all__ = ['PolygonPixelRegion', 'PolygonSkyRegion']


class PolygonPixelRegion(PixelRegion):
    """
    A polygon in pixel coordinates.

    Parameters
    ----------
    vertices : `~regions.PixCoord`
        The vertices of the polygon.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, PolygonPixelRegion
        import matplotlib.pyplot as plt

        x, y = [45, 45, 55, 60], [75, 70, 65, 75]
        fig, ax = plt.subplots(1, 1)

        vertices = PixCoord(x=x, y=y)
        reg = PolygonPixelRegion(vertices=vertices)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(30, 80)
        plt.ylim(50, 80)
        ax.set_aspect('equal')
    """

    _params = ('vertices',)
    vertices = OneDPix('vertices')

    def __init__(self, vertices, meta=None, visual=None):
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        # See https://stackoverflow.com/questions/24467972

        # Use offsets to improve numerical precision
        x_ = self.vertices.x - self.vertices.x.mean()
        y_ = self.vertices.y - self.vertices.y.mean()

        # Shoelace formula, for our case where the start vertex
        # isn't duplicated at the end, written to avoid an array copy
        area_main = np.dot(x_[:-1], y_[1:]) - np.dot(y_[:-1], x_[1:])
        area_last = x_[-1] * y_[0] - y_[-1] * x_[0]
        return 0.5 * np.abs(area_main + area_last)

    def contains(self, pixcoord):
        pixcoord = PixCoord._validate(pixcoord, 'pixcoord')
        x = np.atleast_1d(np.asarray(pixcoord.x, dtype=float))
        y = np.atleast_1d(np.asarray(pixcoord.y, dtype=float))
        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        shape = x.shape
        mask = points_in_polygon(x.flatten(), y.flatten(), vx, vy).astype(bool)
        in_poly = mask.reshape(shape)
        if self.meta.get('include', True):
            return in_poly
        else:
            return np.logical_not(in_poly)

    def to_sky(self, wcs):
        vertices_sky = pixel_to_skycoord(self.vertices.x, self.vertices.y, wcs)
        return PolygonSkyRegion(vertices=vertices_sky)

    @property
    def bounding_box(self):
        xmin = self.vertices.x.min()
        xmax = self.vertices.x.max()
        ymin = self.vertices.y.min()
        ymax = self.vertices.y.max()
        return BoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        if mode == 'subpixels':
            use_exact = 0
        else:
            use_exact = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at
        # origin
        xmin = float(bbox.ixmin) - 0.5
        xmax = float(bbox.ixmax) - 0.5
        ymin = float(bbox.iymin) - 0.5
        ymax = float(bbox.iymax) - 0.5

        vx = np.asarray(self.vertices.x, dtype=float)
        vy = np.asarray(self.vertices.y, dtype=float)

        fraction = polygonal_overlap_grid(xmin, xmax, ymin, ymax, nx, ny,
                                          vx, vy, use_exact, subpixels)

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Polygon`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Polygon`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Polygon`
            A matplotlib polygon patch.
        """
        from matplotlib.patches import Polygon

        xy = np.vstack([self.vertices.x - origin[0],
                        self.vertices.y - origin[1]]).transpose()

        mpl_kwargs = self._define_mpl_kwargs(artist='Patch')
        mpl_kwargs.update(kwargs)

        return Polygon(xy=xy, **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Postive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `PolygonPixelRegion`
            The rotated region (which is an independent copy).
        """
        vertices = self.vertices.rotate(center, angle)
        return self.copy(vertices=vertices)


class PolygonSkyRegion(SkyRegion):
    """
    A polygon defined using vertices in sky coordinates.

    Parameters
    ----------
    vertices : `~astropy.coordinates.SkyCoord`
        The vertices of the polygon.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('vertices',)
    vertices = OneDSky('vertices')

    def __init__(self, vertices, meta=None, visual=None):
        self.vertices = vertices
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        x, y = skycoord_to_pixel(self.vertices, wcs)
        vertices_pix = PixCoord(x, y)
        return PolygonPixelRegion(vertices_pix)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines line regions in both pixel and sky coordinates.
"""

from astropy.wcs.utils import pixel_to_skycoord, skycoord_to_pixel
import numpy as np

from ..core.attributes import ScalarPix, ScalarSky
from ..core.bounding_box import BoundingBox
from ..core.core import PixelRegion, SkyRegion
from ..core.metadata import RegionMeta, RegionVisual
from ..core.pixcoord import PixCoord

__all__ = ['LinePixelRegion', 'LineSkyRegion']


class LinePixelRegion(PixelRegion):
    """
    A line in pixel coordinates.

    Parameters
    ----------
    start : `~regions.PixCoord`
        The start position.
    end : `~regions.PixCoord`
        The end position.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from regions import PixCoord, LinePixelRegion
        import matplotlib.pyplot as plt

        x1, y1 = 15, 10
        x2, y2 = 20, 25

        fig, ax = plt.subplots(1, 1)

        start = PixCoord(x=x1, y=y1)
        end = PixCoord(x=x2, y=y2)
        reg = LinePixelRegion(start=start, end=end)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(0, 30)
        plt.ylim(0, 30)
        ax.set_aspect('equal')
    """

    _params = ('start', 'end')
    start = ScalarPix('start')
    end = ScalarPix('end')

    def __init__(self, start, end, meta=None, visual=None):
        self.start = start
        self.end = end
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return 0

    def contains(self, pixcoord):
        if pixcoord.isscalar:
            in_reg = False
        else:
            in_reg = np.zeros(pixcoord.x.shape, dtype=bool)

        if self.meta.get('include', True):
            return in_reg
        else:
            return np.logical_not(in_reg)

    def to_sky(self, wcs):
        start = pixel_to_skycoord(self.start.x, self.start.y, wcs)
        end = pixel_to_skycoord(self.end.x, self.end.y, wcs)
        return LineSkyRegion(start, end)

    @property
    def bounding_box(self):
        xmin = min(self.start.x, self.end.x)
        xmax = max(self.start.x, self.end.x)
        ymin = min(self.start.y, self.end.y)
        ymax = max(self.start.y, self.end.y)

        return BoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        # TODO: needs to be implemented
        raise NotImplementedError

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Arrow`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Arrow`. These keywords will override
            any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Arrow`
            A matplotlib line patch.
        """
        # Note: Long term we want to support DS9 lines with arrow heads.
        # We may want to use Line2D instead of arrow for lines because the
        # width of the arrow is non-scalable in patches.
        from matplotlib.patches import Arrow

        x = self.start.x - origin[0]
        y = self.start.y - origin[1]
        dx = self.end.x - self.start.x
        dy = self.end.y - self.start.y
        kwargs.setdefault("width", 0.1)

        mpl_kwargs = self._define_mpl_kwargs(artist='Patch')
        mpl_kwargs.update(kwargs)

        return Arrow(x, y, dx, dy, **mpl_kwargs)

    def rotate(self, center, angle):
        """
        Rotate the region.

        Postive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `LinePixelRegion`
            The rotated region (which is an independent copy).
        """
        start = self.start.rotate(center, angle)
        end = self.end.rotate(center, angle)
        return self.copy(start=start, end=end)


class LineSkyRegion(SkyRegion):
    """
    A line in sky coordinates.

    Parameters
    ----------
    start : `~astropy.coordinates.SkyCoord`
        The start position.
    end : `~astropy.coordinates.SkyCoord`
        The end position.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('start', 'end')
    start = ScalarSky('start')
    end = ScalarSky('end')

    def __init__(self, start, end, meta=None, visual=None):
        self.start = start
        self.end = end
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def contains(self, skycoord, wcs):  # pylint: disable=unused-argument
        if self.meta.get('include', True):
            # lines never contain anything
            return False
        else:
            return True

    def to_pixel(self, wcs):
        start_x, start_y = skycoord_to_pixel(self.start, wcs=wcs)
        start = PixCoord(start_x, start_y)
        end_x, end_y = skycoord_to_pixel(self.end, wcs=wcs)
        end = PixCoord(end_x, end_y)
        return LinePixelRegion(start, end)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import abc
import copy
import operator

from .metadata import RegionMeta, RegionVisual
from .pixcoord import PixCoord
from .registry import RegionsRegistry

__all__ = ['Region', 'PixelRegion', 'SkyRegion']
__doctest_skip__ = ['Region.serialize', 'Region.write']


class Region(abc.ABC):
    """
    Base class for all regions.
    """

    _params = ()

    def copy(self, **changes):
        """
        Make an independent (deep) copy.
        """
        fields = list(self._params) + ['meta', 'visual']

        for field in fields:
            if field not in changes:
                changes[field] = copy.deepcopy(getattr(self, field))

        return self.__class__(**changes)

    def __repr__(self):
        prefix = f'{self.__class__.__name__}'
        cls_info = []
        if self._params is not None:
            for param in self._params:
                cls_info.append(f'{param}={getattr(self, param)}')
        cls_info = ', '.join(cls_info)
        return f'<{prefix}({cls_info})>'

    def __str__(self):
        cls_info = [('Region', self.__class__.__name__)]
        if self._params is not None:
            for param in self._params:
                cls_info.append((param, getattr(self, param)))
        return '\n'.join([f'{key}: {val}' for key, val in cls_info])

    @abc.abstractmethod
    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.
        """
        raise NotImplementedError

    def __and__(self, other):
        return self.intersection(other)

    def __or__(self, other):
        return self.union(other)

    def __xor__(self, other):
        return self.symmetric_difference(other)

    @classmethod
    def get_formats(cls):
        """
        Get the registered I/O formats as a Table.
        """
        return RegionsRegistry.get_formats(Region)

    def write(self, filename, format=None, overwrite=False, **kwargs):
        """
        Write the region to a region file in the specified format.

        This method allows writing a file in many supported data
        formats, e.g.,::

            >>> reg.write('new_regions.reg', format='ds9')
            >>> reg.write('new_regions.crtf', format='crtf')
            >>> reg.write('new_regions.fits', format='fits')

        A list of the available formats for `~regions.Region` is
        available using::

            >>> from regions import Region
            >>> Region.get_formats()

        Parameters
        ----------
        filename : str
            The filename or URL of the file to write.

        format : str, optional
            The file format specifier.

        overwrite : bool, optional
            If True, overwrite the output file if it exists. Raises an
            `OSError` if False and the output file exists. Default is
            False.

        **kwargs : dict, optional
            Keyword arguments passed to the data writer.
        """
        return RegionsRegistry.write([self], filename,
                                     Region, format=format,
                                     overwrite=overwrite, **kwargs)

    def serialize(self, format=None, **kwargs):
        """
        Serialize the region to a region string or table.

        This method allows serializing regions in many supported data
        formats, e.g.,::

            >>> reg1_str = reg.serialize(format='ds9')
            >>> reg2_str = reg.serialize(format='crtf')
            >>> reg3_tbl = reg.serialize(format='fits')

        A list of the available formats for `~regions.Region` is
        available using::

            >>> from regions import Region
            >>> Region.get_formats()

        Parameters
        ----------
        format : str, optional
            The file format specifier.

        **kwargs : dict, optional
            Keyword arguments passed to the data serializer.
        """
        return RegionsRegistry.serialize([self], Region, format=format,
                                         **kwargs)


class PixelRegion(Region):
    """
    Base class for all regions defined in pixel coordinates.
    """

    meta = RegionMeta()
    visual = RegionVisual()

    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.xor)

    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.
        """
        from .compound import CompoundPixelRegion
        return CompoundPixelRegion(region1=self, region2=other,
                                   operator=operator.or_)

    @abc.abstractmethod
    def contains(self, pixcoord):
        """
        Check whether a position or positions fall inside the region.

        Parameters
        ----------
        pixcoord : `~regions.PixCoord`
            The position or positions to check.
        """
        raise NotImplementedError

    def __contains__(self, coord):
        if not coord.isscalar:
            raise ValueError(f'coord must be scalar. coord={coord}')
        return self.contains(coord)

    @abc.abstractmethod
    def to_sky(self, wcs):
        """
        Return a region defined in sky coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            from pixels to sky coordinates.

        Returns
        -------
        sky_region : `~regions.SkyRegion`
            The sky region.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def area(self):
        """
        The area of the region.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def bounding_box(self):
        """
        The minimal bounding box (in integer pixel coordinates) that
        contains the region.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_mask(self, mode='center', subpixels=5):
        """
        Return a mask for the region.

        Parameters
        ----------
        mode : {'center', 'exact', 'subpixels'}, optional
            The method used to determine the overlap of the region on
            the pixel grid. Not all options are available for all region
            types. Note that the more precise methods are generally
            slower. The following methods are available:

                * ``'center'``:
                  A pixel is considered to be entirely in or out of the
                  region depending on whether its center is in or out of
                  the region. The returned mask will contain values only
                  of 0 (out) and 1 (in).

                * ``'exact'`` (default):
                  The exact fractional overlap of the region and each
                  pixel is calculated. The returned mask will contain
                  values between 0 and 1.

                * ``'subpixel'``:
                  A pixel is divided into subpixels (see the
                  ``subpixels`` keyword), each of which are considered
                  to be entirely in or out of the region depending
                  on whether its center is in or out of the region.
                  If ``subpixels=1``, this method is equivalent to
                  ``'center'``. The returned mask will contain values
                  between 0 and 1.

        subpixels : int, optional
            For the ``'subpixel'`` mode, resample pixels by this factor
            in each dimension. That is, each pixel is divided into
            ``subpixels ** 2`` subpixels.

        Returns
        -------
        mask : `~regions.RegionMask`
            A mask for the region.
        """
        raise NotImplementedError

    @staticmethod
    def _validate_mode(mode, subpixels):
        valid_modes = ('center', 'exact', 'subpixels')
        if mode not in valid_modes:
            raise ValueError(f'Invalid mask mode: {mode} (should be one '
                             f'of {valid_modes}')

        if mode == 'subpixels':
            if not isinstance(subpixels, int) or subpixels <= 0:
                raise ValueError(f'Invalid subpixels value: {subpixels} '
                                 '(should be a strictly positive integer)')

    @abc.abstractmethod
    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Convert to matplotlib patch object for this region.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch`
            A matplotlib patch.
        """
        raise NotImplementedError

    def _define_mpl_kwargs(self, artist='Patch'):
        """
        Define a dictionary of matplotlib keywords for the input
        ``artist`` from the region's ``visual`` properties.

        Unset ``visual`` properties are set to default values based on
        the DS9 convention.
        """
        if artist not in ('Patch', 'Line2D', 'Text'):
            raise ValueError(f'artist "{artist}" is not supported')

        kwargs = {}
        kwargs['color'] = self.visual.get('color', 'green')
        kwargs['label'] = self.meta.get('label', '')

        if artist == 'Text':
            kwargs['family'] = self.visual.get('font', 'helvetica')
            kwargs['style'] = self.visual.get('fontsyle', 'normal')
            kwargs['weight'] = self.visual.get('fontweight', 'roman')
            kwargs['size'] = self.visual.get('fontsize', '12')
            kwargs['rotation'] = self.visual.get('textangle', '0')
        else:
            kwargs['linewidth'] = self.visual.get('linewidth', 1)
            kwargs['linestyle'] = self.visual.get('linstyle', 'solid')

            if artist == 'Line2D':
                kwargs['marker'] = self.visual.get('symbol', 'o')
                kwargs['markersize'] = self.visual.get('symsize', 11)
                kwargs['markeredgecolor'] = kwargs['color']
                kwargs['markeredgewidth'] = self.visual.get('width', 1)

            if artist == 'Patch':
                kwargs['edgecolor'] = kwargs.pop('color')
                kwargs['fill'] = self.visual.get('fill', False)

        return kwargs

    def plot(self, origin=(0, 0), ax=None, **kwargs):
        """
        Plot the region on a matplotlib `~matplotlib.axes.Axes`
        instance.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        ax : `~matplotlib.axes.Axes` or `None`, optional
            The matplotlib axes on which to plot.  If `None`, then the
            current `~matplotlib.axes.Axes` instance is used.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Patch`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            The axes on which the patch was added.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            ax = plt.gca()

        artist = self.as_artist(origin=origin, **kwargs)
        ax.add_artist(artist)

        return ax


class SkyRegion(Region):
    """
    Base class for all regions defined in celestial coordinates.
    """

    def intersection(self, other):
        """
        Return a region representing the intersection of this region
        with ``other``.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.and_)

    def symmetric_difference(self, other):
        """
        Return the union of the two regions minus any areas contained in
        the intersection of the two regions.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.xor)

    def union(self, other):
        """
        Return a region representing the union of this region with
        ``other``.
        """
        from .compound import CompoundSkyRegion
        return CompoundSkyRegion(region1=self, region2=other,
                                 operator=operator.or_)

    def contains(self, skycoord, wcs):
        """
        Check whether a sky coordinate falls inside the region.

        Parameters
        ----------
        skycoord : `~astropy.coordinates.SkyCoord`
            The position or positions to check.
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            between sky and pixel coordinates.
        """
        pixel_region = self.to_pixel(wcs)
        pixcoord = PixCoord.from_sky(skycoord, wcs)
        return pixel_region.contains(pixcoord)

    @abc.abstractmethod
    def to_pixel(self, wcs):
        """
        Return the equivalent region defined in pixel coordinates.

        Parameters
        ----------
        wcs : `~astropy.wcs.WCS`
            The world coordinate system transformation to use to convert
            between sky and pixel coordinates.

        Returns
        -------
        pixel_region : `~regions.PixelRegion`
            A pixel region.
        """
        raise NotImplementedError



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module proves classes to handle region metadata.
"""

__all__ = ['Meta', 'RegionMeta', 'RegionVisual']


class Meta(dict):
    """
    A base class for region metadata.
    """

    valid_keys = []
    key_mapping = {}

    def __init__(self, seq=None, **kwargs):
        super().__init__()

        if seq:
            if isinstance(seq, dict):
                for key, val in seq.items():
                    self.__setitem__(key, val)
            else:
                for key, val in seq:
                    self.__setitem__(key, val)

        if len(kwargs) > 0:
            for key, val in kwargs.items():
                self.__setitem__(key, val)

    def __setitem__(self, key, value):
        key = self.key_mapping.get(key, key)
        if key in self.valid_keys:
            super().__setitem__(key, value)
        else:
            raise KeyError(f'{key} is not a valid key for this class.')

    def __getitem__(self, item):
        item = self.key_mapping.get(item, item)
        return super().__getitem__(item)


class RegionMeta(Meta):
    """
    A dictionary subclass that holds the meta attributes of the region.
    """

    valid_keys = ['background', 'comment', 'corr', 'delete', 'edit', 'fixed',
                  'frame', 'highlite', 'include', 'label', 'line', 'move',
                  'name', 'range', 'restfreq', 'rotate', 'select', 'source',
                  'tag', 'text', 'type', 'veltype']

    key_mapping = {}


class RegionVisual(Meta):
    """
    A dictionary subclass which holds the visual attributes of the
    region.
    """

    valid_keys = ['color', 'dash', 'dashlist', 'fill', 'font', 'fontsize',
                  'fontstyle', 'fontweight', 'labeloff', 'labelpos', 'line',
                  'linestyle', 'linewidth', 'symbol', 'symsize', 'symthick',
                  'textangle', 'usetex']

    key_mapping = {'point': 'symbol', 'width': 'linewidth'}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

from astropy.utils.decorators import deprecated

from ...core import Region, Regions
from ...core.registry import RegionsRegistry
from ..core import _to_shape_list

__all__ = ['write_crtf', 'crtf_objects_to_string']


@RegionsRegistry.register(Region, 'serialize', 'crtf')
@RegionsRegistry.register(Regions, 'serialize', 'crtf')
def _serialize_crtf(regions, coordsys='fk5', fmt='.6f', radunit='deg'):
    shapelist = _to_shape_list(regions, coordsys)
    return shapelist.to_crtf(coordsys, fmt, radunit)


@deprecated('0.5', alternative='`regions.Regions.serialize`')
def crtf_objects_to_string(regions, coordsys='fk5', fmt='.6f', radunit='deg'):
    """
    Convert a list of `~regions.Region` objects to a CRTF region string.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    coordsys : str, optional
        An Astropy coordinate system that overrides the coordinate
        system frame for all regions.

    fmt : str, optional
        A python string format defining the output precision. Default is
        '.6f', which is accurate to 0.0036 arcseconds.

    radunit : str, optional
        The unit of the radius.

    Returns
    -------
    region_string : str
        A CRTF region string.
    """
    return _serialize_crtf(regions, coordsys=coordsys, fmt=fmt,
                           radunit=radunit)


@RegionsRegistry.register(Region, 'write', 'crtf')
@RegionsRegistry.register(Regions, 'write', 'crtf')
def _write_crtf(regions, filename, coordsys='fk5', fmt='.6f', radunit='deg',
                overwrite=False):
    """
    Convert a list of `~regions.Region` to a CRTF string and write to a
    file.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    filename : str
        The filename in which the string is to be written.

    coordsys : str, optional
        An Astropy coordinate system that overrides the coordinate
        frames of all regions.

    fmt : str, optional
        A python string format defining the output precision. Default is
        '.6f', which is accurate to 0.0036 arcseconds.

    radunit : str, optional
        The unit of the radius.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    if os.path.lexists(filename) and not overwrite:
        raise OSError(f'{filename} already exists')

    output = _serialize_crtf(regions, coordsys=coordsys, fmt=fmt,
                             radunit=radunit)
    with open(filename, 'w') as fh:
        fh.write(output)


@deprecated('0.5', alternative='`regions.Regions.write`')
def write_crtf(regions, filename, coordsys='fk5', fmt='.6f', radunit='deg',
               overwrite=False):
    """
    Convert a list of `~regions.Region` to a CRTF string and write to a
    file.

    Parameters
    ----------
    regions : list
        A list of `~regions.Region` objects.

    filename : str
        The filename in which the string is to be written.

    coordsys : str, optional
        An Astropy coordinate system that overrides the coordinate
        frames of all regions.

    fmt : str, optional
        A python string format defining the output precision. Default is
        '.6f', which is accurate to 0.0036 arcseconds.

    radunit : str, optional
        The unit of the radius.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Raises an
        `OSError` if False and the output file exists. Default is False.
    """
    return _write_crtf(regions, filename, coordsys=coordsys, fmt=fmt,
                       radunit=radunit, overwrite=overwrite)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Base class for all shape tests.
"""

import numpy as np
from numpy.testing import assert_equal, assert_allclose
import pytest

from ...core.pixcoord import PixCoord


class BaseTestRegion:

    def test_repr(self):
        assert repr(self.reg) == self.expected_repr

    def test_str(self):
        assert str(self.reg) == self.expected_str


class BaseTestPixelRegion(BaseTestRegion):

    def test_area(self):
        assert_allclose(self.reg.area, self.expected_area)

    def test_mask_area(self):
        try:
            mask = self.reg.to_mask(mode='exact')
            assert_allclose(np.sum(mask.data), self.expected_area)
        except NotImplementedError:
            try:
                mask = self.reg.to_mask(mode='subpixels', subpixels=30)
                assert_allclose(np.sum(mask.data), self.expected_area,
                                rtol=0.005)
            except NotImplementedError:
                pytest.skip()

    def test_contains_scalar(self):
        if len(self.inside) > 0:
            pixcoord = PixCoord(*self.inside[0])
            assert self.reg.contains(pixcoord)
            assert pixcoord in self.reg

        if len(self.outside) > 0:
            pixcoord = PixCoord(*self.outside[0])
            assert not self.reg.contains(pixcoord)
            assert pixcoord not in self.reg

    def test_contains_array_1d(self):
        pixcoord = PixCoord(*zip(*(self.inside + self.outside)))
        actual = self.reg.contains(pixcoord)
        assert_equal(actual[:len(self.inside)], True)
        assert_equal(actual[len(self.inside):], False)

        with pytest.raises(ValueError) as excinfo:
            pixcoord in self.reg
        assert 'coord must be scalar' in str(excinfo.value)

    def test_contains_array_2d(self):
        x, y = zip(*(self.inside + self.outside))
        pixcoord = PixCoord([x] * 3, [y] * 3)

        actual = self.reg.contains(pixcoord)
        assert actual.shape == (3, len(x))
        assert_equal(actual[:, :len(self.inside)], True)
        assert_equal(actual[:, len(self.inside):], False)


class BaseTestSkyRegion(BaseTestRegion):
    # TODO: here we should add inside/outside tests as above
    pass



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage provides the core functionality for regions.
"""

from .attributes import *  # noqa
from .bounding_box import *  # noqa
from .compound import *  # noqa
from .core import *  # noqa
from .mask import *  # noqa
from .metadata import *  # noqa
from .pixcoord import *  # noqa
from .regions import *  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
import itertools
import re
import string
from warnings import warn

from astropy.coordinates import Angle, frame_transform_graph
import astropy.units as u
from astropy.utils.data import get_readable_fileobj
from astropy.utils.decorators import deprecated

from ...core import Regions
from ...core.registry import RegionsRegistry
from ..core import _Shape, _ShapeList, reg_mapping
from .core import (DS9RegionParserError, DS9RegionParserWarning,
                   valid_symbols_ds9)

__all__ = ['read_ds9', 'DS9Parser']

# Regular expression to extract region type or coordinate system
regex_global = re.compile('^#? *(-?)([a-zA-Z0-9]+)')

# Regular expression to extract meta attributes
regex_meta = re.compile(r'([a-zA-Z]+)(\s*)(=)(\s*)({.*?}|\'.*?\'|\".*?\"|[0-9\s]+\s?|[^=\s]+\s?[0-9]*)\s?')

# Regular expression to strip parenthesis
regex_paren = re.compile('[()]')

# Regular expression to split coordinate strings
regex_splitter = re.compile('[, ]')


@RegionsRegistry.register(Regions, 'read', 'ds9')
def _read_ds9(filename, errors='strict', cache=False):
    """
    Read a DS9 region file in as a list of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The filename of the file to access.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.DS9RegionParserError`. 'warn' will raise a
        `~regions.DS9RegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    with get_readable_fileobj(filename, cache=cache) as fh:
        region_string = fh.read()
        return _parse_ds9(region_string, errors=errors)


@RegionsRegistry.register(Regions, 'parse', 'ds9')
def _parse_ds9(region_string, errors='strict'):
    parser = _DS9Parser(region_string, errors=errors)
    return Regions(parser.shapes.to_regions())


class _DS9CoordinateParser:
    """
    Helper class to structure coordinate parser.
    """

    @staticmethod
    def parse_coordinate(string_rep, unit):
        """
        Parse a single coordinate.
        """
        # explicit radian ('r') value
        if string_rep[-1] == 'r':
            return Angle(string_rep[:-1], unit=u.rad)

        # explicit image ('i') and physical ('p') pixels
        elif string_rep[-1] in ['i', 'p']:
            return u.Quantity(string_rep[:-1]) - 1

        # Any ds9 coordinate representation (sexagesimal or degrees)
        elif 'd' in string_rep or 'h' in string_rep:
            return Angle(string_rep)

        elif unit == 'hour_or_deg':
            if ':' in string_rep:
                spl = tuple(float(x) for x in string_rep.split(':'))
                return Angle(spl, u.hourangle)
            else:
                ang = float(string_rep)
                return Angle(ang, u.deg)

        elif unit.is_equivalent(u.deg):
            # return Angle(string_rep, unit=unit)
            if ':' in string_rep:
                ang = tuple(float(x) for x in string_rep.split(':'))
            else:
                ang = float(string_rep)
            return Angle(ang, u.deg)

        elif unit.is_equivalent(u.dimensionless_unscaled):
            return u.Quantity(float(string_rep), unit) - 1

        else:
            return u.Quantity(float(string_rep), unit)

    @staticmethod
    def parse_angular_length_quantity(string_rep, unit=u.deg):
        """
        Parse a string into a Quantity object.

        Given a string that is a number and a unit, return a Quantity of
        that string, e.g.:

        * 23.9 -> 23.9 * u.deg
        * 50" -> 50 * u.arcsec
        """
        unit_mapping = {'"': u.arcsec,
                        "'": u.arcmin,
                        'd': u.deg,
                        'r': u.rad,
                        'i': u.dimensionless_unscaled,
                        'p': u.dimensionless_unscaled}
        has_unit = string_rep[-1] not in string.digits
        if has_unit:
            unit = unit_mapping[string_rep[-1]]
            return u.Quantity(float(string_rep[:-1]), unit=unit)
        else:
            return u.Quantity(float(string_rep), unit=unit)


class _DS9Parser:
    """
    Parse a DS9 string.

    This class transforms a DS9 string to a
    `~regions.io.core.ShapeList`. The result is stored as ``shapes``
    attribute.

    Parameters
    ----------
    region_string : str
        A DS9 region string.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.DS9RegionParserError`. 'warn' will raise a
        `~regions.DS9RegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """

    # Each line is tested for either containing a region type or a
    # coordinate system. If a coordinate system is found the global
    # coordsys state of the parser is modified. If a region type is
    # found the ``_DS9RegionParser`` is invokes to transform the line
    # into a `~regions.Shape` object.

    # List of valid coordinate system (all lowercase)
    coordinate_systems = ['fk5', 'fk4', 'icrs', 'galactic', 'wcs',
                          'physical', 'image', 'ecliptic', 'j2000']
    coordinate_systems += [f'wcs{letter}' for letter in string.ascii_lowercase]

    # Map to convert coordinate system names
    coordsys_mapping = dict(zip(frame_transform_graph.get_names(),
                                frame_transform_graph.get_names()))
    coordsys_mapping['ecliptic'] = 'geocentrictrueecliptic'
    coordsys_mapping['j2000'] = 'fk5'

    def __init__(self, region_string, errors='strict'):
        if errors not in ('strict', 'ignore', 'warn'):
            raise ValueError('errors must be one of "strict", "ignore", '
                             'or "warn"')
        self.region_string = region_string
        self.errors = errors

        # Global states
        self.coordsys = None
        self.global_meta = {}

        # Results
        self.shapes = _ShapeList()

        self.run()

    def __str__(self):
        ss = self.__class__.__name__
        ss += f'\nErrors: {self.errors}'
        ss += f'\nCoordsys: {self.coordsys}'
        ss += f'\nGlobal meta: {self.global_meta}'
        ss += f'\nShapes: {self.shapes}'
        ss += '\n'
        return ss

    def set_coordsys(self, coordsys):
        """
        Transform coordinate system
        """
        # TODO: needs expert attention
        if coordsys in self.coordsys_mapping:
            self.coordsys = self.coordsys_mapping[coordsys]
        else:
            self.coordsys = coordsys

    def run(self):
        """
        Run all the steps.
        """
        for line_ in self.region_string.split('\n'):
            # split on all semicolons, except those between {} braces
            # (ds9 text strings)
            for line in _split_semicolon(line_):
                self.parse_line(line)

    def parse_line(self, line):
        """
        Parse a single line.
        """
        # Skip blanks
        if line == '':
            return

        # Skip comments
        if line[0] == '#':
            return

        # Special case / header: parse global parameters into metadata
        if line.lstrip()[:6].lower() == 'global':
            self.global_meta = self.parse_meta(line.lower())
            # global_meta can specify "include=1"; never seen other options
            # used but presumably 0 means false
            include = not (self.global_meta.get('include')
                           in ('0', 'False', False))
            self.global_meta['include'] = include

            return

        # Try to parse the line
        region_type_search = regex_global.search(line.lower())
        if region_type_search:
            include = region_type_search.groups()[0]
            region_type = region_type_search.groups()[1]
        else:
            self._raise_error(f'No region type found for line "{line}".')
            return

        if region_type in self.coordinate_systems:
            # Found coord system definition
            self.set_coordsys(region_type)
            return
        if region_type not in _DS9RegionParser.language_spec:
            self._raise_error(f'Region type "{region_type}" was found, but '
                              'it is not one of the supported region types.')
            return
        else:
            # Found region specification,
            region_end = region_type_search.span()[1]
            self.parse_region(include, region_type, region_end, line)

    def _raise_error(self, msg):
        if self.errors == 'warn':
            warn(msg, DS9RegionParserWarning)
        elif self.errors == 'strict':
            raise DS9RegionParserError(msg)

    @staticmethod
    def parse_meta(meta_str):
        """
        Parse the metadata for a single DS9 region string.

        Parameters
        ----------
        meta_str : str
            Meta string, the metadata is everything after the closing
            parenthesis of the region coordinate specification. All
            metadata is specified as key=value pairs separated by
            whitespace, but sometimes the values can also be whitespace
            separated.

        Returns
        -------
        meta : dict
            A dictionary containing the metadata.
        """
        # keys must be lower-case, but data can be any case
        keys_vals = [(x.lower(), y) for x, _, _, _, y
                     in regex_meta.findall(meta_str.strip())]
        extra_text = regex_meta.split(meta_str.strip())[-1]
        result = {}
        for key, val in keys_vals:
            # regex can include trailing whitespace or inverted commas
            # remove it
            val = val.strip().strip("'").strip('"')
            if key == 'text':
                val = val.lstrip('{').rstrip('}')
            if key in result:
                if key == 'tag':
                    result[key].append(val)
                else:
                    raise ValueError(f'Duplicate key {key} found')
            else:
                if key == 'tag':
                    result[key] = [val]
                else:
                    result[key] = val
        if extra_text:
            result['comment'] = extra_text

        return result

    def parse_region(self, include, region_type, region_end, line):
        """
        Extract a Shape from a region string.
        """
        if self.coordsys is None:
            raise DS9RegionParserError('No coordinate system specified and a '
                                       'region has been found.')

        helper = _DS9RegionParser(coordsys=self.coordsys, include=include,
                                  region_type=region_type,
                                  region_end=region_end,
                                  global_meta=self.global_meta, line=line)
        helper.parse()
        self.shapes.append(helper.shape)


# Global definitions to improve readability
radius = _DS9CoordinateParser.parse_angular_length_quantity
width = _DS9CoordinateParser.parse_angular_length_quantity
height = _DS9CoordinateParser.parse_angular_length_quantity
angle = _DS9CoordinateParser.parse_angular_length_quantity
coordinate = _DS9CoordinateParser.parse_coordinate


class _DS9RegionParser:
    """
    Parse a DS9 region string.

    This will turn a line containing a DS9 region into a
    `~regions.Shape` object.

    Parameters
    ----------
    coordsys : str
        The coordinate system.

    include : {'', '-'}
        Flag at the beginning of the line

    region_type : str
        Region type.

    region_end : int
        Coordinate of the end of the region name. This is passed in
        order to handle whitespace correctly.

    global_meta : dict
        Global metadata.

    line : str
        The line to parse.
    """

    # Coordinate unit transformations
    coordinate_units = {'fk5': ('hour_or_deg', u.deg),
                        'fk4': ('hour_or_deg', u.deg),
                        'icrs': ('hour_or_deg', u.deg),
                        'geocentrictrueecliptic': (u.deg, u.deg),
                        'galactic': (u.deg, u.deg),
                        'physical': (u.dimensionless_unscaled,
                                     u.dimensionless_unscaled),
                        'image': (u.dimensionless_unscaled,
                                  u.dimensionless_unscaled),
                        'wcs': (u.dimensionless_unscaled,
                                u.dimensionless_unscaled)}

    for letter in string.ascii_lowercase:
        coordinate_units[f'wcs{letter}'] = (u.dimensionless_unscaled,
                                            u.dimensionless_unscaled)

    # DS9 language specification. This defines how a certain region is read.
    language_spec = {'point': (coordinate, coordinate),
                     'text': (coordinate, coordinate),
                     'circle': (coordinate, coordinate, radius),
                     # This is a special case to deal with n elliptical annuli
                     'ellipse': itertools.chain((coordinate, coordinate),
                                                itertools.cycle((radius,))),
                     'box': (coordinate, coordinate, width, height, angle),
                     'polygon': itertools.cycle((coordinate,)),
                     'line': (coordinate, coordinate, coordinate, coordinate),
                     'annulus': itertools.chain((coordinate, coordinate),
                                                itertools.cycle((radius,)))}

    def __init__(self, coordsys, include, region_type, region_end,
                 global_meta, line):
        self.coordsys = coordsys
        self.include = include
        self.region_type = region_type
        self.region_end = region_end
        self.global_meta = global_meta
        self.line = line

        self.meta_str = None
        self.coord_str = None
        self.composite = None
        self.coord = None
        self.meta = None
        self.shape = None

    def __str__(self):
        ss = self.__class__.__name__
        ss += f'\nLine : {self.line}'
        ss += f'\nRegion end : {self.region_end}'
        ss += f'\nMeta string : {self.meta_str}'
        ss += f'\nCoord string: {self.coord_str}'
        ss += f'\nShape: {self.shape}'
        ss += '\n'
        return ss

    def parse(self):
        """
        Parse the region string.
        """
        self.parse_composite()
        self.split_line()
        self.convert_coordinates()
        self.convert_meta()
        self.make_shape()

    def parse_composite(self):
        """
        Determine whether the region is composite.
        """
        self.composite = '||' in self.line

    def split_line(self):
        """
        Split line into coordinates and meta string.
        """
        # index of the # symbol or end of the line (-1) if not found
        hash_idx = self.line.find('#')
        if hash_idx == -1:
            temp = self.line[self.region_end:].strip(' |')
            self.meta_str = ''  # no metadata found
        else:
            temp = self.line[self.region_end:hash_idx].strip(' |')
            self.meta_str = self.line[hash_idx:]

        # force all coordinate names (circle, etc) to be lower-case
        self.coord_str = regex_paren.sub('', temp).lower()

    def convert_coordinates(self):
        """
        Convert coordinate string to objects.
        """
        coord_list = []
        # strip out 'null' elements, i.e., ''. It might be possible to
        # eliminate these some other way, i.e., with regex directly.
        # We need to copy in order not to burn up the iterators.
        elements = [x for x in regex_splitter.split(self.coord_str) if x]
        element_parsers = self.language_spec[self.region_type]
        for ii, (element, element_parser) in enumerate(zip(elements,
                                                           element_parsers)):
            if element_parser is coordinate:
                unit = self.coordinate_units[self.coordsys][ii % 2]
                coord_list.append(element_parser(element, unit))
            elif (self.coordinate_units[self.coordsys][0]
                  is u.dimensionless_unscaled):
                coord_list.append(
                    element_parser(element, unit=u.dimensionless_unscaled))
            else:
                coord_list.append(element_parser(element))

        if self.region_type in ['ellipse', 'box'] and len(coord_list) % 2 == 1:
            coord_list[-1] = _DS9CoordinateParser.parse_angular_length_quantity(
                elements[len(coord_list) - 1])

        # Reset iterator for ellipse and annulus
        # Note that this cannot be done with copy.deepcopy on python2
        if self.region_type in ['ellipse', 'annulus']:
            self.language_spec[self.region_type] = itertools.chain(
                (coordinate, coordinate), itertools.cycle((radius,)))

        self.coord = coord_list

    def convert_meta(self):
        """
        Convert meta string to dict.
        """
        meta_ = DS9Parser.parse_meta(self.meta_str)
        self.meta = copy.deepcopy(self.global_meta)
        self.meta.update(meta_)
        # the 'include' is not part of the metadata string;
        # it is pre-parsed as part of the shape type and should always
        # override the global one
        self.include = (self.meta.get('include', True)
                        if self.include == '' else self.include != '-')
        self.meta['include'] = self.include

    def make_shape(self):
        """
        Make shape object.
        """
        # In DS9, ellipse can also represents an elliptical annulus
        # For elliptical annulus angle is optional.
        if self.region_type == 'ellipse':
            self.coord[2:] = [x * 2 for x in self.coord[2:]]
            if len(self.coord) % 2 == 1:  # This checks if angle is present
                self.coord[-1] /= 2

        if 'point' in self.meta:
            point = self.meta['point'].split(' ')
            if len(point) > 1:
                self.meta['symsize'] = point[1]
            self.meta['point'] = valid_symbols_ds9[point[0]]

        if 'font' in self.meta:
            fonts = self.meta['font'].split(' ')
            keys = ['font', 'fontsize', 'fontstyle', 'fontweight']
            for i, val in enumerate(fonts):
                self.meta[keys[i]] = val

        self.meta.pop('coord', None)

        self.shape = _Shape(coordsys=self.coordsys,
                            region_type=reg_mapping['DS9'][self.region_type],
                            coord=self.coord, meta=self.meta,
                            composite=self.composite, include=self.include)


def _find_text_delim_idx(regstr):
    """
    Find the indices of the DS9 text field delimiters ({}, '', or "") in
    a string.
    """
    pattern = re.compile(r'(text\s*=\s*[{\'"])')
    idx0 = []
    delim = []
    start_idx = []
    for match in pattern.finditer(regstr):
        idx0.append(match.start())
        end_delim = match.group()[-1]
        if end_delim == '{':
            end_delim = '}'
        delim.append(end_delim)
        start_idx.append(match.span()[1])

    idx1 = []
    for sidx, char in zip(start_idx, delim):
        idx1.append(regstr.find(char, sidx))

    return idx0, idx1


def _split_semicolon(regstr):
    r"""
    Split a DS9 region string on semicolons.  The line is not split on
    semicolons found in a text field.

    This turned out to be a very trickly problem (attempts with regex failed)
      * text fields are delimited by {}, '', or ""
      * the text delimiters do not have to be consistent within a file
      * region strings can have unpaired ' (arcmin) or " (arcsec)
      * region strings can have "#" in a color value (cannot split on #)
      * the text field can contain "text={str}", "text='str'",
        'test="str"' etc., e.g., text={text="hello"}
      * the delimiter characters (escaped or not) used by the text field
        are not allowed within the text field, e.g., "text={my
        field\{test\}}" and "text='my field \'test\''" are invalid.
        However, "text={my field, 'test'}" is valid

    This code finds the text field delimiters and then finds the indices
    of the opening and closing delimiters. Text fields that contain
    "text={str}", etc. are included (as a smaller range between the
    larger text field range), but this is fine because all we need are
    index ranges where to exclude semicolons (for splitting). Semicolons
    found at indices between the open/close delimiter indices are
    excluded from splitting.
    """
    idx0, idx1 = _find_text_delim_idx(regstr)

    semi_idx = [pos for pos, char in enumerate(regstr) if char == ';']
    fidx = []
    for i in semi_idx:
        for i0, i1 in zip(idx0, idx1):
            if i >= i0 and i <= i1:
                break
        else:
            fidx.append(i + 1)
    fidx.insert(0, 0)

    return [regstr[i:j].rstrip(';') for i, j in zip(fidx, fidx[1:] + [None])]


@deprecated('0.5', alternative='`regions.Regions.read`')
def read_ds9(filename, errors='strict', cache=False):
    """
    Read a DS9 region file in as a list of `~regions.Region` objects.

    Parameters
    ----------
    filename : str
        The filename of the file to access.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.DS9RegionParserError`. 'warn' will raise a
        `~regions.DS9RegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).

    cache : bool or 'update', optional
        Whether to cache the contents of remote URLs. If 'update', check
        the remote URL for a new version but store the result in the
        cache.

    Returns
    -------
    regions : list
        A list of `~regions.Region` objects.
    """
    return _read_ds9(filename, errors=errors, cache=cache)


@deprecated('0.5', alternative='`regions.Regions.parse`')
class DS9Parser(_DS9Parser):
    """
    Parse a DS9 string.

    This class transforms a DS9 string to a
    `~regions.io.core.ShapeList`. The result is stored as ``shapes``
    attribute.

    Parameters
    ----------
    region_string : str
        A DS9 region string.

    errors : {'strict', 'warn', 'ignore'}, optional
        The error handling scheme to use for handling parsing
        errors. The default is 'strict', which will raise a
        `~regions.DS9RegionParserError`. 'warn' will raise a
        `~regions.DS9RegionParserWarning`, and 'ignore' will do nothing
        (i.e., be silent).
    """



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This module defines rectangular regions in both pixel and sky coordinates.
"""

from astropy.coordinates import Angle
import astropy.units as u
from astropy.wcs.utils import pixel_to_skycoord
import numpy as np

from ..core.attributes import (ScalarPix, ScalarLength, QuantityLength,
                               ScalarSky)
from ..core.bounding_box import BoundingBox
from ..core.core import PixelRegion, SkyRegion
from ..core.mask import RegionMask
from ..core.metadata import RegionMeta, RegionVisual
from ..core.pixcoord import PixCoord
from .._geometry import rectangular_overlap_grid
from .._utils.wcs_helpers import pixel_scale_angle_at_skycoord

from .polygon import PolygonPixelRegion

__all__ = ['RectanglePixelRegion', 'RectangleSkyRegion']


class RectanglePixelRegion(PixelRegion):
    """
    A rectangle in pixel coordinates.

    Parameters
    ----------
    center : `~regions.PixCoord`
        The position of the center of the rectangle.
    width : float
        The width of the rectangle (before rotation) in pixels.
    height : float
        The height of the rectangle (before rotation) in pixels.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangle, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the x
        axis.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.

    Examples
    --------
    .. plot::
        :include-source:

        from astropy.coordinates import Angle
        from regions import PixCoord, RectanglePixelRegion
        import matplotlib.pyplot as plt

        x, y = 15, 10
        width, height = 8, 5
        angle = Angle(30, 'deg')

        fig, ax = plt.subplots(1, 1)

        center = PixCoord(x=x, y=y)
        reg = RectanglePixelRegion(center=center, width=width,
                                   height=height, angle=angle)
        patch = reg.as_artist(facecolor='none', edgecolor='red', lw=2)
        ax.add_patch(patch)

        plt.xlim(0, 30)
        plt.ylim(0, 20)
        ax.set_aspect('equal')
    """

    _params = ('center', 'width', 'height', 'angle')
    center = ScalarPix('center')
    width = ScalarLength('width')
    height = ScalarLength('height')
    angle = QuantityLength('angle')

    def __init__(self, center, width, height, angle=0 * u.deg, meta=None,
                 visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    @property
    def area(self):
        return self.width * self.height

    def contains(self, pixcoord):
        cos_angle = np.cos(self.angle)
        sin_angle = np.sin(self.angle)
        dx = pixcoord.x - self.center.x
        dy = pixcoord.y - self.center.y
        dx_rot = cos_angle * dx + sin_angle * dy
        dy_rot = sin_angle * dx - cos_angle * dy
        in_rect = ((np.abs(dx_rot) < self.width * 0.5)
                   & (np.abs(dy_rot) < self.height * 0.5))
        if self.meta.get('include', True):
            return in_rect
        else:
            return np.logical_not(in_rect)

    def to_sky(self, wcs):
        center = pixel_to_skycoord(self.center.x, self.center.y, wcs)
        _, pixscale, north_angle = pixel_scale_angle_at_skycoord(center, wcs)
        width = Angle(self.width * u.pix * pixscale, 'arcsec')
        height = Angle(self.height * u.pix * pixscale, 'arcsec')
        angle = self.angle - (north_angle - 90 * u.deg)
        return RectangleSkyRegion(center, width, height, angle=angle,
                                  meta=self.meta, visual=self.visual)

    @property
    def bounding_box(self):
        w2 = self.width / 2.
        h2 = self.height / 2.
        cos_angle = np.cos(self.angle)  # self.angle is a Quantity
        sin_angle = np.sin(self.angle)  # self.angle is a Quantity
        dx1 = abs(w2 * cos_angle - h2 * sin_angle)
        dy1 = abs(w2 * sin_angle + h2 * cos_angle)
        dx2 = abs(w2 * cos_angle + h2 * sin_angle)
        dy2 = abs(w2 * sin_angle - h2 * cos_angle)
        dx = max(dx1, dx2)
        dy = max(dy1, dy2)

        xmin = self.center.x - dx
        xmax = self.center.x + dx
        ymin = self.center.y - dy
        ymax = self.center.y + dy

        return BoundingBox.from_float(xmin, xmax, ymin, ymax)

    def to_mask(self, mode='center', subpixels=5):
        # NOTE: assumes this class represents a single rectangle

        self._validate_mode(mode, subpixels)

        if mode == 'center':
            mode = 'subpixels'
            subpixels = 1

        # Find bounding box and mask size
        bbox = self.bounding_box
        ny, nx = bbox.shape

        # Find position of pixel edges and recenter so that circle is at
        # origin
        xmin = float(bbox.ixmin) - 0.5 - self.center.x
        xmax = float(bbox.ixmax) - 0.5 - self.center.x
        ymin = float(bbox.iymin) - 0.5 - self.center.y
        ymax = float(bbox.iymax) - 0.5 - self.center.y

        if mode == 'subpixels':
            use_exact = 0
        else:
            use_exact = 1

        fraction = rectangular_overlap_grid(xmin, xmax, ymin, ymax, nx, ny,
                                            self.width, self.height,
                                            self.angle.to(u.rad).value,
                                            use_exact, subpixels)

        return RegionMask(fraction, bbox=bbox)

    def as_artist(self, origin=(0, 0), **kwargs):
        """
        Return a matplotlib patch object for this region
        (`matplotlib.patches.Rectangle`).

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` pixel position of the origin of the displayed
            image.

        **kwargs : dict
            Any keyword arguments accepted by
            `~matplotlib.patches.Rectangle`. These keywords will
            override any visual meta attributes of this region.

        Returns
        -------
        artist : `~matplotlib.patches.Rectangle`
            A matplotlib rectangle patch.
        """
        from matplotlib.patches import Rectangle

        xy = self._lower_left_xy()
        xy = xy[0] - origin[0], xy[1] - origin[1]
        width = self.width
        height = self.height
        # matplotlib expects rotation in degrees (anti-clockwise)
        angle = self.angle.to('deg').value

        mpl_kwargs = self._define_mpl_kwargs(artist='Patch')
        mpl_kwargs.update(kwargs)

        return Rectangle(xy=xy, width=width, height=height,
                         angle=angle, **mpl_kwargs)

    def _update_from_mpl_selector(self, *args, **kwargs):
        xmin, xmax, ymin, ymax = self._mpl_selector.extents
        self.center = PixCoord(x=0.5 * (xmin + xmax),
                               y=0.5 * (ymin + ymax))
        self.width = (xmax - xmin)
        self.height = (ymax - ymin)
        self.angle = 0. * u.deg
        if self._mpl_selector_callback is not None:
            self._mpl_selector_callback(self)

    def as_mpl_selector(self, ax, active=True, sync=True, callback=None,
                        **kwargs):
        """
        A matplotlib editable widget for this region
        (`matplotlib.widgets.RectangleSelector`).

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes to add the selector to.
        active : bool, optional
            Whether the selector should be active by default.
        sync : bool, optional
            If `True` (the default), the region will be kept in
            sync with the selector. Otherwise, the selector will be
            initialized with the values from the region but the two will
            then be disconnected.
        callback : callable, optional
            If specified, this function will be called every time the
            region is updated. This only has an effect if ``sync`` is
            `True`. If a callback is set, it is called for the first
            time once the selector has been created.
        **kwargs : dict
            Additional keyword arguments are passed to
            `matplotlib.widgets.RectangleSelector`.

        Returns
        -------
        selector : `matplotlib.widgets.RectangleSelector`
            The matplotlib selector.

        Notes
        -----
        Once a selector has been created, you will need to keep a
        reference to it until you no longer need it. In addition,
        you can enable/disable the selector at any point by calling
        ``selector.set_active(True)`` or ``selector.set_active(False)``.
        """
        from matplotlib.widgets import RectangleSelector

        if hasattr(self, '_mpl_selector'):
            raise Exception('Cannot attach more than one selector to a '
                            'region.')

        if self.angle.value != 0:
            raise NotImplementedError('Cannot create matplotlib selector for '
                                      'rotated rectangle.')

        if sync:
            sync_callback = self._update_from_mpl_selector
        else:
            def sync_callback(*args, **kwargs):
                pass

        self._mpl_selector = RectangleSelector(
            ax, sync_callback, interactive=True,
            rectprops={'edgecolor': self.visual.get('color', 'black'),
                       'facecolor': 'none',
                       'linewidth': self.visual.get('linewidth', 1),
                       'linestyle': self.visual.get('linestyle', 'solid')})

        self._mpl_selector.extents = (self.center.x - self.width / 2,
                                      self.center.x + self.width / 2,
                                      self.center.y - self.height / 2,
                                      self.center.y + self.height / 2)
        self._mpl_selector.set_active(active)
        self._mpl_selector_callback = callback

        if sync and self._mpl_selector_callback is not None:
            self._mpl_selector_callback(self)

        return self._mpl_selector

    @property
    def corners(self):
        """
        Return the x, y coordinate pairs that define the corners.
        """
        corners = [(-self.width / 2, -self.height / 2),
                   (self.width / 2, -self.height / 2),
                   (self.width / 2, self.height / 2),
                   (-self.width / 2, self.height / 2),
                   ]
        rotmat = [[np.cos(self.angle), np.sin(self.angle)],
                  [-np.sin(self.angle), np.cos(self.angle)]]

        return np.dot(corners, rotmat) + np.array([self.center.x,
                                                   self.center.y])

    def to_polygon(self):
        """
        Return a 4-sided polygon equivalent to this rectangle.
        """
        x, y = self.corners.T
        vertices = PixCoord(x=x, y=y)
        return PolygonPixelRegion(vertices=vertices, meta=self.meta,
                                  visual=self.visual)

    def _lower_left_xy(self):
        """
        Compute lower left ``xy`` pixel position.

        This is used for the conversion to matplotlib in ``as_artist``.

        Taken from
        http://photutils.readthedocs.io/en/latest/_modules/photutils/ape
        rture/rectangle.html#RectangularAperture.plot.
        """
        hw = self.width / 2.
        hh = self.height / 2.
        sint = np.sin(self.angle)
        cost = np.cos(self.angle)
        dx = (hh * sint) - (hw * cost)
        dy = -(hh * cost) - (hw * sint)
        x = self.center.x + dx
        y = self.center.y + dy
        return x, y

    def rotate(self, center, angle):
        """
        Rotate the region.

        Postive ``angle`` corresponds to counter-clockwise rotation.

        Parameters
        ----------
        center : `~regions.PixCoord`
            The rotation center point.
        angle : `~astropy.coordinates.Angle`
            The rotation angle.

        Returns
        -------
        region : `RectanglePixelRegion`
            The rotated region (which is an independent copy).
        """
        center = self.center.rotate(center, angle)
        angle = self.angle + angle
        return self.copy(center=center, angle=angle)


class RectangleSkyRegion(SkyRegion):
    """
    A rectangle in sky coordinates.

    Parameters
    ----------
    center : `~astropy.coordinates.SkyCoord`
        The position of the center of the rectangle.
    width : `~astropy.units.Quantity`
        The width of the rectangle (before rotation) as an angle.
    height : `~astropy.units.Quantity`
        The height of the rectangle (before rotation) as an angle.
    angle : `~astropy.units.Quantity`, optional
        The rotation angle of the rectangle, measured anti-clockwise. If
        set to zero (the default), the width axis is lined up with the
        longitude axis of the celestial coordinates.
    meta : `~regions.RegionMeta`, optional
        A dictionary that stores the meta attributes of this region.
    visual : `~regions.RegionVisual`, optional
        A dictionary that stores the visual meta attributes of this
        region.
    """

    _params = ('center', 'width', 'height', 'angle')
    center = ScalarSky('center')
    width = QuantityLength('width')
    height = QuantityLength('height')
    angle = QuantityLength('angle')

    def __init__(self, center, width, height, angle=0 * u.deg, meta=None,
                 visual=None):
        self.center = center
        self.width = width
        self.height = height
        self.angle = angle
        self.meta = meta or RegionMeta()
        self.visual = visual or RegionVisual()

    def to_pixel(self, wcs):
        center, pixscale, north_angle = pixel_scale_angle_at_skycoord(
            self.center, wcs)
        width = (self.width / pixscale).to(u.pix).value
        height = (self.height / pixscale).to(u.pix).value
        angle = self.angle + (north_angle - 90 * u.deg)
        return RectanglePixelRegion(center, width, height, angle=angle,
                                    meta=self.meta, visual=self.visual)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.utils.data import get_readable_fileobj

from ...core import Region, Regions
from ...core.registry import RegionsRegistry

__all__ = []


@RegionsRegistry.register(Region, 'identify', 'ds9')
@RegionsRegistry.register(Regions, 'identify', 'ds9')
def is_ds9(methodname, filepath):
    """
    Identify a DS9 region file.

    Parameters
    ----------
    methodname : {'read', 'write'}
        The method name called that needs auto-identification.

    filepath : str
        The path to the file.

    Returns
    -------
    result : bool
        Returns `True` if the given file is a DS9 region file.
    """
    all_exten = ('.ds9', '.reg', '.ds9.gz', '.reg.gz')
    exten = {'read': all_exten, 'write': all_exten[0:2]}

    if methodname == 'write':
        return filepath.lower().endswith(exten[methodname])

    elif methodname == 'read':
        if (isinstance(filepath, str)
                and filepath.lower().endswith(exten[methodname])):
            return True
        else:
            with get_readable_fileobj(filepath, encoding='binary') as fileobj:
                signature = '# Region file format: DS9'
                pos = fileobj.tell()
                sig = fileobj.read(len(signature))
                fileobj.seek(pos)
                return sig == signature or sig == signature.encode()

    else:
        return False
